import pytest
import numpy as np
import numpy.testing as npt

import lenstronomy.Util.util as util

from lenstronomy.Data.psf import PSF
from lenstronomy.Data.kinematic_bin_2D import KinBin

import lenstronomy.Util.kernel_util as kernel_util


class TestKinBin(object):
    def setup(self):
        self.numPix = 10
        self.num_bin = 4
        kwargs_kin = {
            "bin_data": np.zeros(self.num_bin) + 10.0,
            "bin_cov": np.diag(np.ones(self.num_bin) * 2.0**2),
            "bin_mask": np.zeros((self.numPix, self.numPix)),
        }
        kwargs_kin["bin_mask"][0, 0] = 1
        kwargs_kin["bin_mask"][0, 1] = 1
        kwargs_kin["bin_mask"][1:5, 2:4] = 2
        kwargs_kin["bin_mask"][9, 9] = 3

        kwargs_kin["bin_data"][1] = 20.0
        kwargs_kin["transform_pix2angle"] = np.array([[1, 0], [0, 1]])
        kwargs_kin["ra_at_xy_0"] = 0
        kwargs_kin["dec_at_xy_0"] = 0

        kernel_point_source = kernel_util.kernel_gaussian(
            num_pix=9, delta_pix=1.0, fwhm=2.0
        )
        kwargs_pixel = {"psf_type": "PIXEL", "kernel_point_source": kernel_point_source}
        self.PSF = PSF(**kwargs_pixel)
        self.kwargs_kin = kwargs_kin
        self.KinBin = KinBin(psf_class=self.PSF, **kwargs_kin)

    def test_binned_image(self):
        assert (
            self.KinBin.binned_image(
                self.kwargs_kin["bin_data"], self.kwargs_kin["bin_mask"]
            )[0, 0]
            == 20.0
        )

    def test_KinBin2kwargs(self):
        assert self.KinBin.kin_bin2kwargs()["deltaPix"] == 1.0

    def test_kin_grid(self):
        assert np.shape(self.KinBin.kin_grid()[0]) == (self.numPix, self.numPix)



__all__ = ["PointSourceCached"]


class PointSourceCached(object):
    """This class is the same as PointSource() except that it saves image and source
    positions in cache.

    This speeds-up repeated calls for the same source and lens model and avoids duplicating the lens equation solving.
    Attention: cache needs to be deleted before calling functions with different lens and point source parameters.
    """

    def __init__(self, point_source_model, save_cache=False):
        self._model = point_source_model
        self._save_cache = save_cache

    def delete_lens_model_cache(self):
        if hasattr(self, "_x_image"):
            del self._x_image
        if hasattr(self, "_y_image"):
            del self._y_image
        if hasattr(self, "_x_image_add"):
            del self._x_image_add
        if hasattr(self, "_y_image_add"):
            del self._y_image_add
        if hasattr(self, "_x_source"):
            del self._x_source
        if hasattr(self, "_y_source"):
            del self._y_source

    def set_save_cache(self, save_bool):
        self._save_cache = save_bool

    def update_lens_model(self, lens_model_class):
        self._model.update_lens_model(lens_model_class)

    def image_position(
        self,
        kwargs_ps,
        kwargs_lens=None,
        magnification_limit=None,
        kwargs_lens_eqn_solver=None,
        additional_images=False,
    ):
        """On-sky image positions.

        :param kwargs_ps: keyword arguments of the point source model
        :param kwargs_lens: keyword argument list of the lens model(s), only used when
            requiring the lens equation solver
        :param magnification_limit: float >0 or None, if float is set and additional
            images are computed, only those images will be computed that exceed the
            lensing magnification (absolute value) limit
        :param kwargs_lens_eqn_solver: keyword arguments specifying the numerical
            settings for the lens equation solver see LensEquationSolver() class for
            details
        :param additional_images: if True, solves the lens equation for additional
            images
        :type additional_images: bool
        :return: image positions in x, y as arrays
        """
        if additional_images and not self._model.additional_images:
            # ignore cached parts if additional images
            if (
                not self._save_cache
                or not hasattr(self, "_x_image_add")
                or not hasattr(self, "_y_image_add")
            ):
                self._x_image_add, self._y_image_add = self._model.image_position(
                    kwargs_ps,
                    kwargs_lens=kwargs_lens,
                    magnification_limit=magnification_limit,
                    kwargs_lens_eqn_solver=kwargs_lens_eqn_solver,
                    additional_images=additional_images,
                )
            return self._x_image_add, self._y_image_add
        if (
            not self._save_cache
            or not hasattr(self, "_x_image")
            or not hasattr(self, "_y_image")
        ):
            self._x_image, self._y_image = self._model.image_position(
                kwargs_ps,
                kwargs_lens=kwargs_lens,
                magnification_limit=magnification_limit,
                kwargs_lens_eqn_solver=kwargs_lens_eqn_solver,
                additional_images=additional_images,
            )
        return self._x_image, self._y_image

    def source_position(self, kwargs_ps, kwargs_lens=None):
        """Original source position (prior to lensing)

        :param kwargs_ps: point source keyword arguments
        :param kwargs_lens: lens model keyword argument list (only used when required)
        :return: x, y position
        """
        if (
            not self._save_cache
            or not hasattr(self, "_x_source")
            or not hasattr(self, "_y_source")
        ):
            self._x_source, self._y_source = self._model.source_position(
                kwargs_ps, kwargs_lens=kwargs_lens
            )
        return self._x_source, self._y_source

    def image_amplitude(
        self,
        kwargs_ps,
        kwargs_lens=None,
        magnification_limit=None,
        kwargs_lens_eqn_solver=None,
    ):
        """Image brightness amplitudes.

        :param kwargs_ps: keyword arguments of the point source model
        :param kwargs_lens: keyword argument list of the lens model(s), only used when
            requiring the lens equation solver
        :param magnification_limit: float >0 or None, if float is set and additional
            images are computed, only those images will be computed that exceed the
            lensing magnification (absolute value) limit
        :param kwargs_lens_eqn_solver: keyword arguments specifying the numerical
            settings for the lens equation solver see LensEquationSolver() class for
            details
        :return: array of image amplitudes
        """
        x_pos, y_pos = self.image_position(
            kwargs_ps,
            kwargs_lens,
            magnification_limit=magnification_limit,
            kwargs_lens_eqn_solver=kwargs_lens_eqn_solver,
        )
        return self._model.image_amplitude(
            kwargs_ps, kwargs_lens=kwargs_lens, x_pos=x_pos, y_pos=y_pos
        )

    def source_amplitude(self, kwargs_ps, kwargs_lens=None):
        """Intrinsic brightness amplitude of point source.

        :param kwargs_ps: keyword arguments of the point source model
        :param kwargs_lens: keyword argument list of the lens model(s), only used when
            positions are defined in image plane and have to be ray-traced back
        :return: brightness amplitude (as numpy array)
        """
        return self._model.source_amplitude(kwargs_ps, kwargs_lens=kwargs_lens)



from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
from lenstronomy.Util import param_util
from lenstronomy.Util import derivative_util
import numpy as np

__all__ = ["ArcPerturbations"]


class ArcPerturbations(LensProfileBase):
    """Uses radial and tangential fourier modes within a specific range in both
    directions to perturb a lensing potential."""

    def __init__(self):
        super(ArcPerturbations, self).__init__()
        self._2_pi = np.pi * 2

    def function(self, x, y, coeff, d_r, d_phi, center_x, center_y):
        """

        :param x: x-coordinate
        :param y: y-coordinate
        :param coeff: float, amplitude of basis
        :param d_r: period of radial sinusoidal in units of angle
        :param d_phi: period of tangential sinusoidal in radian
        :param center_x: center of rotation for tangential basis
        :param center_y: center of rotation for tangential basis
        :return:
        """
        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        dphi_ = d_phi / self._2_pi
        phi_r = self._phi_r(r, d_r)
        phi_theta = self._phi_theta(phi, dphi_)
        return phi_r * phi_theta * coeff

    def derivatives(self, x, y, coeff, d_r, d_phi, center_x, center_y):
        """

        :param x: x-coordinate
        :param y: y-coordinate
        :param coeff: float, amplitude of basis
        :param d_r: period of radial sinusoidal in units of angle
        :param d_phi: period of tangential sinusoidal in radian
        :param center_x: center of rotation for tangential basis
        :param center_y: center of rotation for tangential basis
        :return: f_x, f_y
        """
        r, theta = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        dphi_ = d_phi / self._2_pi
        d_phi_dr = self._d_phi_r(r, d_r) * self._phi_theta(theta, dphi_)
        d_phi_d_theta = self._d_phi_theta(theta, dphi_) * self._phi_r(r, d_r)
        x_ = x - center_x
        y_ = y - center_y
        dr_dx = derivative_util.d_r_dx(x_, y_)
        dr_dy = derivative_util.d_r_dy(x_, y_)
        d_theta_dx = derivative_util.d_phi_dx(x_, y_)
        d_theta_dy = derivative_util.d_phi_dy(x_, y_)
        f_x = d_phi_dr * dr_dx + d_phi_d_theta * d_theta_dx
        f_y = d_phi_dr * dr_dy + d_phi_d_theta * d_theta_dy
        return f_x * coeff, f_y * coeff

    def hessian(self, x, y, coeff, d_r, d_phi, center_x, center_y):
        """

        :param x: x-coordinate
        :param y: y-coordinate
        :param coeff: float, amplitude of basis
        :param d_r: period of radial sinusoidal in units of angle
        :param d_phi: period of tangential sinusoidal in radian
        :param center_x: center of rotation for tangential basis
        :param center_y: center of rotation for tangential basis
        :return: f_xx, f_yy, f_xy
        """
        r, theta = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        dphi_ = d_phi / self._2_pi

        d_phi_dr = self._d_phi_r(r, d_r) * self._phi_theta(theta, dphi_)
        d_phi_dr2 = self._d_phi_r2(r, d_r) * self._phi_theta(theta, dphi_)
        d_phi_d_theta = self._d_phi_theta(theta, dphi_) * self._phi_r(r, d_r)
        d_phi_d_theta2 = self._d_phi_theta2(theta, dphi_) * self._phi_r(r, d_r)
        d_phi_dr_dtheta = self._d_phi_r(r, d_r) * self._d_phi_theta(theta, dphi_)
        x_ = x - center_x
        y_ = y - center_y
        dr_dx = derivative_util.d_r_dx(x_, y_)
        dr_dy = derivative_util.d_r_dy(x_, y_)
        d_theta_dx = derivative_util.d_phi_dx(x_, y_)
        d_theta_dy = derivative_util.d_phi_dy(x_, y_)
        dr_dxx = derivative_util.d_x_diffr_dx(x_, y_)
        dr_dxy = derivative_util.d_x_diffr_dy(x_, y_)
        dr_dyy = derivative_util.d_y_diffr_dy(x_, y_)
        d_theta_dxx = derivative_util.d_phi_dxx(x_, y_)
        d_theta_dyy = derivative_util.d_phi_dyy(x_, y_)
        d_theta_dxy = derivative_util.d_phi_dxy(x_, y_)

        f_xx = (
            d_phi_dr2 * dr_dx**2
            + d_phi_dr * dr_dxx
            + d_phi_d_theta2 * d_theta_dx**2
            + d_phi_d_theta * d_theta_dxx
            + 2 * d_phi_dr_dtheta * dr_dx * d_theta_dx
        )
        f_yy = (
            d_phi_dr2 * dr_dy**2
            + d_phi_dr * dr_dyy
            + d_phi_d_theta2 * d_theta_dy**2
            + d_phi_d_theta * d_theta_dyy
            + 2 * d_phi_dr_dtheta * dr_dy * d_theta_dy
        )
        f_xy = (
            d_phi_dr2 * dr_dx * dr_dy
            + d_phi_dr * dr_dxy
            + d_phi_d_theta2 * d_theta_dx * d_theta_dy
            + d_phi_d_theta * d_theta_dxy
            + d_phi_dr_dtheta * dr_dx * d_theta_dy
            + d_phi_dr_dtheta * dr_dy * d_theta_dx
        )
        return f_xx * coeff, f_xy * coeff, f_xy * coeff, f_yy * coeff

    @staticmethod
    def _phi_r(r, d_r):
        """

        :param r: numpy array, radius
        :param d_r: period of radial sinusoidal in units of angle
        :return: radial component of the potential
        """
        return np.cos(r / d_r)

    @staticmethod
    def _d_phi_r(r, d_r):
        """Radial derivatives.

        :param r: numpy array, radius
        :param d_r: period of radial sinusoidal in units of angle
        :return: radial component of the potential
        """
        return -np.sin(r / d_r) / d_r

    @staticmethod
    def _d_phi_r2(r, d_r):
        """Radial second derivatives.

        :param r: numpy array, radius
        :param d_r: period of radial sinusoidal in units of angle
        :return: radial component of the potential
        """
        return -np.cos(r / d_r) / d_r**2

    @staticmethod
    def _phi_theta(theta, d_theta):
        """

        :param theta: numpy array, orientation angle in 2pi convention
        :param d_theta: period of tangential sinusoidal in radian
        :return: tangential component of the potential
        """
        return np.cos(theta / d_theta)

    @staticmethod
    def _d_phi_theta(theta, d_theta):
        """Tangential derivatives.

        :param theta: numpy array, angle
        :param d_theta: period of tangential sinusoidal in radian
        :return: tangential component of the potential
        """
        return -np.sin(theta / d_theta) / d_theta

    @staticmethod
    def _d_phi_theta2(r, d_theta):
        """Tangential derivatives.

        :param r: numpy array, radius
        :param d_theta: period of tangential sinusoidal in radian
        :return: tangential component of the potential
        """
        return -np.cos(r / d_theta) / d_theta**2



__author__ = "nataliehogg"

import numpy as np
import numpy.testing as npt
import pytest
import unittest

from lenstronomy.LensModel.single_plane import SinglePlane
from lenstronomy.LensModel.LineOfSight.single_plane_los import SinglePlaneLOS
from lenstronomy.LensModel.MultiPlane.multi_plane import MultiPlane
from lenstronomy.LensModel.Profiles.sis import SIS
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver

from astropy.cosmology import default_cosmology

cosmo = default_cosmology.get()

try:
    import fastell4py

    bool_test = True
except:
    bool_test = False


class TestSinglePlaneLOS(object):
    """Tests the SinglePlaneLOS routines.

    these functions are the same as in TestLensModel but with the addition of LOS and
    LOS_MINIMAL as profiles. with all params in self.kwargs_los set to zero, the results
    should be the same as the non-LOS cases originally tested the test_los_vs_multiplane
    checks that a multiplane setup with three shear planes returns the same as the LOS
    and LOS MINIMAL models
    """

    def setup_method(self):
        self.lensModel = SinglePlane(["GAUSSIAN"])
        self.lensModel_los = SinglePlaneLOS(["GAUSSIAN", "LOS"], index_los=1)
        self.lensModel_minimal = SinglePlaneLOS(
            ["GAUSSIAN", "LOS_MINIMAL"], index_los=1
        )
        self.kwargs = {
            "amp": 1.0,
            "sigma_x": 2.0,
            "sigma_y": 2.0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        self.los_kwargs = {
            "gamma1_os": 0.0,
            "gamma2_os": 0.0,
            "kappa_os": 0.0,
            "omega_os": 0.0,
            "gamma1_od": 0.0,
            "gamma2_od": 0.0,
            "kappa_od": 0.0,
            "omega_od": 0.0,
            "gamma1_ds": 0.0,
            "gamma2_ds": 0.0,
            "kappa_ds": 0.0,
            "omega_ds": 0.0,
            "gamma1_los": 0.0,
            "gamma2_los": 0.0,
            "kappa_los": 0.0,
            "omega_los": 0.0,
        }

    def test_potential(self):
        output = self.lensModel.potential(x=1.0, y=1.0, kwargs=[self.kwargs])
        output_los = self.lensModel_los.potential(
            x=1.0, y=1.0, kwargs=[self.kwargs, self.los_kwargs]
        )
        output_minimal = self.lensModel_minimal.potential(
            x=1.0, y=1.0, kwargs=[self.kwargs, self.los_kwargs]
        )
        npt.assert_almost_equal(output_los, output, decimal=8)
        npt.assert_almost_equal(output_minimal, output, decimal=8)

    def test_alpha(self):
        output1, output2 = self.lensModel.alpha(x=1.0, y=1.0, kwargs=[self.kwargs])
        output1_los, output2_los = self.lensModel_los.alpha(
            x=1.0, y=1.0, kwargs=[self.kwargs, self.los_kwargs]
        )
        output1_minimal, output2_minimal = self.lensModel_minimal.alpha(
            x=1.0, y=1.0, kwargs=[self.kwargs, self.los_kwargs]
        )
        npt.assert_almost_equal(output1_los, output1, decimal=8)
        npt.assert_almost_equal(output2_los, output2, decimal=8)
        npt.assert_almost_equal(output1_minimal, output1, decimal=8)
        npt.assert_almost_equal(output2_minimal, output2, decimal=8)

    def test_ray_shooting(self):
        delta_x, delta_y = self.lensModel.ray_shooting(
            x=1.0, y=1.0, kwargs=[self.kwargs]
        )
        delta_x_los, delta_y_los = self.lensModel_los.ray_shooting(
            x=1.0, y=1.0, kwargs=[self.kwargs, self.los_kwargs]
        )
        delta_x_minimal, delta_y_minimal = self.lensModel_minimal.ray_shooting(
            x=1.0, y=1.0, kwargs=[self.kwargs, self.los_kwargs]
        )
        npt.assert_almost_equal(delta_x_los, delta_x, decimal=8)
        npt.assert_almost_equal(delta_y_los, delta_y, decimal=8)
        npt.assert_almost_equal(delta_x_minimal, delta_x, decimal=8)
        npt.assert_almost_equal(delta_y_minimal, delta_y, decimal=8)

    def test_mass_2d(self):
        mass_kwargs = {"amp": 1.0, "sigma": 2.0, "center_x": 0.0, "center_y": 0.0}

        lensModel = SinglePlane(["GAUSSIAN_KAPPA"])
        lensModel_los = SinglePlaneLOS(["GAUSSIAN_KAPPA", "LOS"], index_los=1)
        lensModel_minimal = SinglePlaneLOS(
            ["GAUSSIAN_KAPPA", "LOS_MINIMAL"], index_los=1
        )

        output = lensModel.mass_2d(r=1, kwargs=[mass_kwargs])
        output_los = lensModel_los.mass_2d(r=1, kwargs=[mass_kwargs, self.los_kwargs])
        output_minimal = lensModel_minimal.mass_2d(
            r=1, kwargs=[mass_kwargs, self.los_kwargs]
        )

        npt.assert_almost_equal(output_los, output, decimal=8)
        npt.assert_almost_equal(output_minimal, output, decimal=8)

    def test_mass_3d(self):
        mass_kwargs = {"amp": 1.0, "sigma": 2.0, "center_x": 0.0, "center_y": 0.0}

        lensModel = SinglePlane(["GAUSSIAN_KAPPA"])
        lensModel_los = SinglePlaneLOS(["GAUSSIAN_KAPPA", "LOS"], index_los=1)
        lensModel_minimal = SinglePlaneLOS(
            ["GAUSSIAN_KAPPA", "LOS_MINIMAL"], index_los=1
        )

        output = lensModel.mass_3d(r=1, kwargs=[mass_kwargs])
        output_los = lensModel_los.mass_3d(r=1, kwargs=[mass_kwargs, self.los_kwargs])
        output_minimal = lensModel_minimal.mass_3d(
            r=1, kwargs=[mass_kwargs, self.los_kwargs]
        )

        npt.assert_almost_equal(output_los, output, decimal=8)
        npt.assert_almost_equal(output_minimal, output, decimal=8)

    def test_density(self):
        theta_E = 1
        r = 1
        sis = SIS()
        density_model = sis.density_lens(r=r, theta_E=theta_E)

        # LOS
        lensModel_los = SinglePlaneLOS(lens_model_list=["SIS", "LOS"], index_los=1)
        density_los = lensModel_los.density(
            r=r, kwargs=[{"theta_E": theta_E}, self.los_kwargs]
        )
        npt.assert_almost_equal(density_los, density_model, decimal=8)

        # LOS_MINIMAL
        lensModel_minimal = SinglePlaneLOS(
            lens_model_list=["SIS", "LOS_MINIMAL"], index_los=1
        )
        density_minimal = lensModel_minimal.density(
            r=r, kwargs=[{"theta_E": theta_E}, self.los_kwargs]
        )
        npt.assert_almost_equal(density_minimal, density_model, decimal=8)

    def test_bool_list(self):
        lensModel_los = SinglePlaneLOS(["SPEP", "SHEAR", "LOS"], index_los=2)
        lensModel_minimal = SinglePlaneLOS(
            ["SPEP", "SHEAR", "LOS_MINIMAL"], index_los=2
        )
        kwargs = [
            {
                "theta_E": 1,
                "gamma": 2,
                "e1": 0.1,
                "e2": -0.1,
                "center_x": 0,
                "center_y": 0,
            },
            {"gamma1": 0.01, "gamma2": -0.02},
            self.los_kwargs,
        ]

        # LOS
        alphax_1_los, alphay_1_los = lensModel_los.alpha(1, 1, kwargs, k=0)
        alphax_1_list, alphay_1_list = lensModel_los.alpha(1, 1, kwargs, k=[0])
        npt.assert_almost_equal(alphax_1_los, alphax_1_list, decimal=5)
        npt.assert_almost_equal(alphay_1_los, alphay_1_list, decimal=5)

        alphax_1_1_los, alphay_1_1_los = lensModel_los.alpha(1, 1, kwargs, k=0)
        alphax_1_2_los, alphay_1_2_los = lensModel_los.alpha(1, 1, kwargs, k=1)
        alphax_full, alphay_full = lensModel_los.alpha(1, 1, kwargs, k=None)
        npt.assert_almost_equal(alphax_1_1_los + alphax_1_2_los, alphax_full, decimal=5)
        npt.assert_almost_equal(alphay_1_1_los + alphay_1_2_los, alphay_full, decimal=5)

        # LOS_MINIMAL
        alphax_1_minimal, alphay_1_minimal = lensModel_minimal.alpha(1, 1, kwargs, k=0)
        alphax_1_list_minimal, alphay_1_list_minimal = lensModel_minimal.alpha(
            1, 1, kwargs, k=[0]
        )
        npt.assert_almost_equal(alphax_1_minimal, alphax_1_list_minimal, decimal=5)
        npt.assert_almost_equal(alphay_1_minimal, alphay_1_list_minimal, decimal=5)

        alphax_1_1_minimal, alphay_1_1_minimal = lensModel_minimal.alpha(
            1, 1, kwargs, k=0
        )
        alphax_1_2_minimal, alphay_1_2_minimal = lensModel_minimal.alpha(
            1, 1, kwargs, k=1
        )
        alphax_full_minimal, alphay_full_minimal = lensModel_minimal.alpha(
            1, 1, kwargs, k=None
        )
        npt.assert_almost_equal(
            alphax_1_1_minimal + alphax_1_2_minimal, alphax_full_minimal, decimal=5
        )
        npt.assert_almost_equal(
            alphay_1_1_minimal + alphay_1_2_minimal, alphay_full_minimal, decimal=5
        )

    def test_los_versus_multiplane(self):
        """This function asserts that the outcome from LOS and LOS MINIMAL is the same
        as MultiPlane."""
        # set up the cosmology to convert between shears
        # the exact numbers don't matter because we are just doing a comparison

        z_o = 0.0  # redshift of observer
        z_d = 0.5  # redshift of main lens
        z_s = 2.0  # redshift of source

        z_f = (z_o + z_d) / 2
        z_b = (z_d + z_s) / 2

        gamma1_od = 0.05
        gamma2_od = -0.01
        gamma1_os = 0.02
        gamma2_os = 0.1
        gamma1_ds = -0.04
        gamma2_ds = 0.03

        def d(z1, z2):
            return cosmo.angular_diameter_distance_z1z2(z1, z2).to_value()

        # conversion of linear LOS shears to lenstronomy convention
        gamma1_f = gamma1_od * (
            (d(z_o, z_d) * d(z_f, z_s)) / (d(z_o, z_s) * d(z_f, z_d))
        )
        gamma2_f = gamma2_od * (
            (d(z_o, z_d) * d(z_f, z_s)) / (d(z_o, z_s) * d(z_f, z_d))
        )

        gamma1_b = gamma1_ds * (
            (d(z_o, z_b) * d(z_d, z_s)) / (d(z_o, z_s) * d(z_d, z_b))
        )
        gamma2_b = gamma2_ds * (
            (d(z_o, z_b) * d(z_d, z_s)) / (d(z_o, z_s) * d(z_d, z_b))
        )

        gamma1_d = gamma1_os - gamma1_f - gamma1_b
        gamma2_d = gamma2_os - gamma2_f - gamma2_b

        # compute non-linear correction to os term
        Identity = np.identity(2)

        Gamma_f = np.array([[gamma1_f, gamma2_f], [gamma2_f, -gamma1_f]])

        Gamma_d = np.array([[gamma1_d, gamma2_d], [gamma2_d, -gamma1_d]])

        Gamma_b = np.array([[gamma1_b, gamma2_b], [gamma2_b, -gamma1_b]])

        Gamma_od = np.array([[gamma1_od, gamma2_od], [gamma2_od, -gamma1_od]])

        Gamma_ofb = np.array(Gamma_f) * (
            (d(z_o, z_s) * d(z_f, z_b)) / (d(z_o, z_b) * d(z_f, z_s))
        )

        Gamma_odb = np.array(Gamma_d) * (
            (d(z_o, z_s) * d(z_d, z_b)) / (d(z_o, z_b) * d(z_d, z_s))
        )

        Gamma_os = (
            Gamma_f
            + Gamma_d
            + Gamma_b
            - np.matmul(Gamma_d, Gamma_od)
            - np.matmul(Gamma_b, Gamma_ofb + np.matmul(Gamma_odb, Identity - Gamma_od))
        )

        kappa_os = (Gamma_os[0, 0] + Gamma_os[1, 1]) / 2
        omega_os = (Gamma_os[1, 0] - Gamma_os[0, 1]) / 2
        gamma1_os = (Gamma_os[0, 0] - Gamma_os[1, 1]) / 2
        gamma2_os = (Gamma_os[0, 1] + Gamma_os[1, 0]) / 2

        # test three image positions
        x, y = np.array([3, 4, 5]), np.array([2, 1, 0])

        lens_model_list = ["EPL", "SHEAR", "SHEAR", "SHEAR"]

        redshift_list = [z_d, z_f, z_d, z_b]

        kwargs_los = {
            "kappa_os": kappa_os,
            "omega_os": omega_os,
            "gamma1_os": gamma1_os,
            "gamma2_os": gamma2_os,
            "kappa_od": 0.0,
            "omega_od": 0.0,
            "gamma1_od": gamma1_od,
            "gamma2_od": gamma2_od,
            "kappa_ds": 0.0,
            "omega_ds": 0.0,
            "gamma1_ds": gamma1_ds,
            "gamma2_ds": gamma2_ds,
        }

        kwargs_epl = {
            "theta_E": 0.8,
            "gamma": 1.95,
            "center_x": 0,
            "center_y": 0,
            "e1": 0.07,
            "e2": -0.03,
        }

        kwargs_gamma_f = {"gamma1": gamma1_f, "gamma2": gamma2_f}
        kwargs_gamma_d = {"gamma1": gamma1_d, "gamma2": gamma2_d}
        kwargs_gamma_b = {"gamma1": gamma1_b, "gamma2": gamma2_b}

        kwargs_singleplane_los = [kwargs_los, kwargs_epl]

        lens_model_los = SinglePlaneLOS(["LOS", "EPL"], index_los=0)

        kwargs_multiplane = [kwargs_epl, kwargs_gamma_f, kwargs_gamma_d, kwargs_gamma_b]

        lens_model_multiplane = MultiPlane(
            z_source=z_s,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
        )
        # set the tolerance
        # ray shooting passes at 1e-16
        # hessian around 1e-6
        tolerance = 1e-5

        # compare some different results from single_plane_los and multiplane
        # we use assert_allclose rather than assert_almost_equal because we are dealing with arrays
        # since we pass an array of image positions

        # displacement angle
        alpha_multiplane_x, alpha_multiplane_y = lens_model_multiplane.alpha(
            x, y, kwargs_multiplane
        )
        alpha_los_x, alpha_los_y = lens_model_los.alpha(x, y, kwargs_singleplane_los)
        npt.assert_allclose(alpha_multiplane_x, alpha_los_x, rtol=tolerance)
        npt.assert_allclose(alpha_multiplane_y, alpha_los_y, rtol=tolerance)

        # ray_shooting
        beta_multiplane_x, beta_multiplane_y = lens_model_multiplane.ray_shooting(
            x, y, kwargs_multiplane
        )
        beta_los_x, beta_los_y = lens_model_los.ray_shooting(
            x, y, kwargs_singleplane_los
        )
        npt.assert_allclose(beta_multiplane_x, beta_los_x, rtol=tolerance)
        npt.assert_allclose(beta_multiplane_y, beta_los_y, rtol=tolerance)

        # hessian
        (
            hessian_multiplane_xx,
            hessian_multiplane_xy,
            hessian_multiplane_yx,
            hessian_multiplane_yy,
        ) = lens_model_multiplane.hessian(x, y, kwargs_multiplane)
        (
            hessian_los_xx,
            hessian_los_xy,
            hessian_los_yx,
            hessian_los_yy,
        ) = lens_model_los.hessian(x, y, kwargs_singleplane_los)
        npt.assert_allclose(hessian_multiplane_xx, hessian_los_xx, rtol=tolerance)
        npt.assert_allclose(hessian_multiplane_xy, hessian_los_xy, rtol=tolerance)
        npt.assert_allclose(hessian_multiplane_yx, hessian_los_yx, rtol=tolerance)
        npt.assert_allclose(hessian_multiplane_yy, hessian_los_yy, rtol=tolerance)

        # time delays
        ra_source, dec_source = 0.05, 0.02
        number_of_images = 4

        lens_model_multiplane_time = LensModel(
            lens_model_list,
            z_lens=z_d,
            z_source=z_s,
            lens_redshift_list=redshift_list,
            multi_plane=True,
        )

        multiplane_solver = LensEquationSolver(lens_model_multiplane_time)
        x_image_mp, y_image_mp = multiplane_solver.findBrightImage(
            ra_source, dec_source, kwargs_multiplane, numImages=number_of_images
        )

        t_days_mp = lens_model_multiplane_time.arrival_time(
            x_image_mp, y_image_mp, kwargs_multiplane
        )
        dt_days_mp = t_days_mp[1:] - t_days_mp[0]

        lens_model_los_time = LensModel(["LOS", "EPL"], z_lens=z_d, z_source=z_s)
        kwargs_time_los = [kwargs_los, kwargs_epl]

        los_solver = LensEquationSolver(lens_model_los_time)
        x_image_los, y_image_los = los_solver.findBrightImage(
            ra_source, dec_source, kwargs_time_los, numImages=number_of_images
        )

        t_days_los = lens_model_los_time.arrival_time(
            x_image_los, y_image_los, kwargs_time_los
        )
        dt_days_los = t_days_los[1:] - t_days_los[0]

        npt.assert_allclose(dt_days_mp, dt_days_los, rtol=tolerance)

    def test_init(self):
        # need to do this for los minimal too?
        lens_model_list = [
            "TNFW",
            "TRIPLE_CHAMELEON",
            "LOS",
            "SHEAR_GAMMA_PSI",
            "CURVED_ARC_CONST",
            "NFW_MC",
            "ARC_PERT",
            "MULTIPOLE",
            "CURVED_ARC_SPP",
        ]
        lensModel = SinglePlaneLOS(lens_model_list=lens_model_list, index_los=0)
        assert lensModel.func_list[0].param_names[0] == "Rs"


class TestRaise(unittest.TestCase):
    def test_raise(self):
        """Check whether raises occurs if fastell4py is not installed.

        :return:
        """
        if bool_test is False:
            with self.assertRaises(ImportError):
                SinglePlaneLOS(lens_model_list=["PEMD", "LOS"], index_los=1)
            with self.assertRaises(ImportError):
                SinglePlaneLOS(lens_model_list=["SPEMD", "LOS"], index_los=1)
        else:
            SinglePlaneLOS(lens_model_list=["PEMD", "SPEMD", "LOS"], index_los=2)


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



__author__ = "sibirrer"

import time
import copy
from lenstronomy.Sampling.Pool.pool import choose_pool
from lenstronomy.Sampling.Likelihoods.image_likelihood import ImageLikelihood
from lenstronomy.Sampling.Samplers.pso import ParticleSwarmOptimizer

__all__ = ["FluxCalibration", "CalibrationLikelihood"]


class FluxCalibration(object):
    """Class to fit coordinate system alignment and flux amplitude calibrations."""

    def __init__(self, kwargs_imaging, kwargs_model, kwargs_params, calibrate_bands):
        """Initialise the classes of the chain and for parameter options for the flux
        calibration fitting.

        :param kwargs_imaging: keyword argument related to imaging data and imaging
            likelihood. Feeds into ImageLikelihood(**kwargs_imaging)
        :param kwargs_model: keyword argument of model components
        :param kwargs_params: keyword argument of model parameters
        :param calibrate_bands: state which bands the flux calibration is applied to
        :type calibrate_bands: list of booleans of length of the imaging bands
        """
        multi_band_list = kwargs_imaging["multi_band_list"]
        multi_band_type = kwargs_imaging["multi_band_type"]

        if calibrate_bands is None:
            calibrate_bands = [False] * len(multi_band_list)
        if multi_band_type != "joint-linear":
            raise ValueError(
                "flux calibration should only be done with join-linear data model!"
            )
        self._calibrate_bands = calibrate_bands
        self.chain = CalibrationLikelihood(
            kwargs_model,
            kwargs_params,
            calibrate_bands=calibrate_bands,
            kwargs_imaging=kwargs_imaging,
        )

    def pso(
        self,
        n_particles=10,
        n_iterations=10,
        threadCount=1,
        mpi=False,
        scaling_lower_limit=0,
        scaling_upper_limit=1000,
        print_key="flux calibration",
    ):
        """Returns the best fit for the lens model on catalogue basis with particle
        swarm optimizer.

        :param n_particles: number of particles in the PSO
        :param n_iterations: number of iterations of the PSO
        :param threadCount: number of threads
        :param mpi: boolean, MPI mode
        :param scaling_lower_limit: lower limit of the flux_scaling initialization
        :param scaling_upper_limit: upper limit of the flux_scaling initialization
        :param print_key: string, print statement
        :return: multi_band_list, [chi2_list, pos_list, vel_list]
        """
        init_pos = self.chain.get_args(self.chain.multi_band_list)
        pool = choose_pool(mpi=mpi, processes=threadCount, use_dill=True)
        num_param = self.chain.num_param
        lower_limit = [scaling_lower_limit] * num_param
        upper_limit = [scaling_upper_limit] * num_param

        pso = ParticleSwarmOptimizer(
            self.chain, lower_limit, upper_limit, n_particles, pool=pool
        )
        pso.set_global_best(
            init_pos, [0] * len(init_pos), self.chain.likelihood(init_pos)
        )

        if pool.is_master():
            print("Computing the %s ..." % print_key)

        time_start = time.time()

        result, [chi2_list, pos_list, vel_list] = pso.optimize(n_iterations)

        multi_band_list = self.chain.update_data(result)

        if pool.is_master():
            time_end = time.time()
            print("parameters found: ", result)
            print(time_end - time_start, "time used for ", print_key)
            print("Calibration completed for bands %s." % self._calibrate_bands)
        return multi_band_list, [chi2_list, pos_list, vel_list]


class CalibrationLikelihood(object):
    def __init__(self, kwargs_model, kwargs_params, calibrate_bands, kwargs_imaging):
        """Initializes all the classes needed for the chain.

        :param kwargs_model: keyword argument of model components
        :param kwargs_params: keyword argument of model parameters
        :param calibrate_bands: state which bands the flux calibration is applied to
        :type calibrate_bands: list of booleans of length of the imaging bands
        :param kwargs_imaging: keyword arguments of the imaging likelihood
        """
        self._calibrate_bands = calibrate_bands
        self._kwargs_model = kwargs_model
        self._kwargs_params = kwargs_params
        self._kwargs_imaging_likelihood = copy.deepcopy(kwargs_imaging)
        self.multi_band_list = self._kwargs_imaging_likelihood["multi_band_list"]

    def _likelihood(self, args):
        """Routine to compute X2 given variable parameters for a MCMC/PSO chainF."""
        # generate image and computes likelihood
        multi_band_list = self.update_data(args)
        self._kwargs_imaging_likelihood["multi_band_list"] = multi_band_list
        # this line is redundant since the self.multi_band_list variable got already updated
        image_likelihood = ImageLikelihood(
            kwargs_model=self._kwargs_model, **self._kwargs_imaging_likelihood
        )
        log_likelihood = image_likelihood.logL(**self._kwargs_params)
        return log_likelihood

    def __call__(self, a):
        return self._likelihood(a)

    def likelihood(self, a):
        return self._likelihood(a)

    def setup(self):
        pass

    def update_data(self, args):
        """

        :param args:
        :return: updated multi_band_list
        """
        k = 0
        for i, band in enumerate(self.multi_band_list):
            if self._calibrate_bands[i]:
                kwargs_data = band[0]
                kwargs_data["flux_scaling"] = args[k]
                k += 1
        return self.multi_band_list

    def get_args(self, multi_band_list):
        """

        :param multi_band_list: list of multi_band [[kwargs_data, kwargs_psf, kwargs_numeric], [...], ...]
        :return:
        """
        args = []
        for i, band in enumerate(multi_band_list):
            if self._calibrate_bands[i]:
                args.append(band[0].get("flux_scaling", 1))
        return args

    @property
    def num_param(self):
        n = 0
        for i in range(len(self._calibrate_bands)):
            if self._calibrate_bands[i]:
                n += 1
        return n



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
import pytest
import lenstronomy.Util.param_util as param_util


class TestSIE(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        from lenstronomy.LensModel.Profiles.sie import SIE
        from lenstronomy.LensModel.Profiles.epl import EPL
        from lenstronomy.LensModel.Profiles.nie import NIE

        self.sie = SIE(NIE=False)
        self.sie_nie = SIE(NIE=True)
        self.epl = EPL()
        self.nie = NIE()

    def test_function(self):
        x = np.array([1])
        y = np.array([2])
        theta_E = 1.0
        q = 0.9
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        values = self.sie.function(x, y, theta_E, e1, e2)
        gamma = 2
        values_spemd = self.epl.function(x, y, theta_E, gamma, e1, e2)
        assert values == values_spemd

        values_nie = self.sie_nie.function(x, y, theta_E, e1, e2)
        s_scale = 0.0000001
        values_spemd = self.nie.function(x, y, theta_E, e1, e2, s_scale)
        npt.assert_almost_equal(values_nie, values_spemd, decimal=6)

    def test_derivatives(self):
        x = np.array([1])
        y = np.array([2])
        theta_E = 1.0
        q = 0.7
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        values = self.sie.derivatives(x, y, theta_E, e1, e2)
        gamma = 2
        values_spemd = self.epl.derivatives(x, y, theta_E, gamma, e1, e2)
        assert values == values_spemd

        values = self.sie_nie.derivatives(x, y, theta_E, e1, e2)
        s_scale = 0.0000001
        values_spemd = self.nie.derivatives(x, y, theta_E, e1, e2, s_scale)
        npt.assert_almost_equal(values, values_spemd, decimal=6)

    def test_hessian(self):
        x = np.array([1])
        y = np.array([2])
        theta_E = 1.0
        q = 0.7
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        values = self.sie.hessian(x, y, theta_E, e1, e2)
        gamma = 2
        values_spemd = self.epl.hessian(x, y, theta_E, gamma, e1, e2)
        assert values[0] == values_spemd[0]

        values = self.sie_nie.hessian(x, y, theta_E, e1, e2)
        s_scale = 0.0000001
        values_spemd = self.nie.hessian(x, y, theta_E, e1, e2, s_scale)
        npt.assert_almost_equal(values, values_spemd, decimal=5)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"


from lenstronomy.LightModel.linear_basis import LinearBasis

__all__ = ["LightModel"]


class LightModel(LinearBasis):
    """Class to handle extended surface brightness profiles (for e.g. source and lens
    light)

    all profiles come with a surface_brightness parameterization (in units per square
    angle and independent of the pixel scale). The parameter 'amp' is the linear scaling
    parameter of surface brightness. Some functional forms come with a total_flux()
    definition that provide the integral of the surface brightness for a given set of
    parameters.

    The SimulationAPI module allows to use astronomical magnitudes to be used and
    translated into the surface brightness conventions of this module given a magnitude
    zero point.
    """

    def __init__(
        self,
        light_model_list,
        deflection_scaling_list=None,
        source_redshift_list=None,
        smoothing=0.001,
        sersic_major_axis=None,
    ):
        """

        :param light_model_list: list of light models
        :param deflection_scaling_list: list of floats indicating a relative scaling of the deflection angle from the
         reduced angles in the lens model definition (optional, only possible in single lens plane with multiple source
         planes)
        :param source_redshift_list: list of redshifts for the different light models
         (optional and only used in multi-plane lensing in conjunction with a cosmology model)
        :param smoothing: smoothing factor for certain models (deprecated)
        :param sersic_major_axis: boolean or None, if True, uses the semi-major axis as the definition of the Sersic
         half-light radius, if False, uses the product average of semi-major and semi-minor axis. If None, uses the
         convention in the lenstronomy yaml setting (which by default is =False)
        """
        super(LightModel, self).__init__(
            light_model_list=light_model_list,
            smoothing=smoothing,
            sersic_major_axis=sersic_major_axis,
        )
        self.deflection_scaling_list = deflection_scaling_list
        self.redshift_list = source_redshift_list



from lenstronomy.LensModel.single_plane import SinglePlane
import numpy as np

__all__ = ["LensParam"]


class LensParam(object):
    """Class to handle the lens model parameter."""

    def __init__(
        self,
        lens_model_list,
        kwargs_fixed,
        kwargs_lower=None,
        kwargs_upper=None,
        kwargs_logsampling=None,
        num_images=0,
        solver_type="NONE",
        num_shapelet_lens=0,
    ):
        """

        :param lens_model_list: list of strings of lens model names
        :param kwargs_fixed: list of keyword arguments for model parameters to be held fixed
        :param kwargs_lower: list of keyword arguments of the lower bounds of the model parameters
        :param kwargs_upper: list of keyword arguments of the upper bounds of the model parameters
        :param kwargs_logsampling: list of keyword arguments of parameters to be sampled in log10 space
        :param num_images: number of images to be constrained by a non-linear solver
         (only relevant when shapelet potential functions are used)
        :param solver_type: string, type of non-linear solver
         (only relevant in this class when 'SHAPELETS' is the solver type)
        :param num_shapelet_lens: integer, number of shapelets in the lensing potential
         (only relevant when 'SHAPELET' lens model is used)
        """

        self.model_list = lens_model_list
        self.kwargs_fixed = kwargs_fixed
        self._num_images = num_images
        self._solver_type = solver_type
        self._num_shapelet_lens = num_shapelet_lens

        # PF: Here the instantiation of SinglePlane is only made to deal with
        # a list of lens parameters. It does not seem to matter whether multi-
        # plane lensing or line-of-sight effects are activated.
        lens_model = SinglePlane(lens_model_list=lens_model_list)

        name_list = []
        for func in lens_model.func_list:
            name_list.append(func.param_names)
        self._param_name_list = name_list
        if kwargs_lower is None:
            kwargs_lower = []
            for func in lens_model.func_list:
                kwargs_lower.append(func.lower_limit_default)
        if kwargs_upper is None:
            kwargs_upper = []
            for func in lens_model.func_list:
                kwargs_upper.append(func.upper_limit_default)

        self.lower_limit = kwargs_lower
        self.upper_limit = kwargs_upper
        if kwargs_logsampling is None:
            kwargs_logsampling = [[] for i in range(len(self.model_list))]
        self.kwargs_logsampling = kwargs_logsampling

    def get_params(self, args, i):
        """

        :param args: tuple of individual floats of sampling argument
        :param i: integer, index at the beginning of the tuple for read out to keyword argument convention
        :return: kwargs_list, index at the end of read out of this model component
        """

        kwargs_list = []
        for k, model in enumerate(self.model_list):
            kwargs = {}
            kwargs_fixed = self.kwargs_fixed[k]
            kwargs_logsampling = self.kwargs_logsampling[k]
            param_names = self._param_name_list[k]
            for name in param_names:
                if name not in kwargs_fixed:
                    if (
                        model in ["SHAPELETS_POLAR", "SHAPELETS_CART"]
                        and name == "coeffs"
                    ):
                        num_coeffs = self._num_shapelet_lens
                        if self._solver_type == "SHAPELETS" and k == 0:
                            if self._num_images == 4:
                                num_coeffs -= 6
                                coeffs = args[i : i + num_coeffs]
                                coeffs = [0, 0, 0, 0, 0, 0] + list(coeffs[0:])
                            elif self._num_images == 2:
                                num_coeffs -= 3
                                coeffs = args[i : i + num_coeffs]
                                coeffs = [0, 0, 0] + list(coeffs[0:])
                            else:
                                raise ValueError("Option for solver_type not valid!")
                            kwargs["coeffs"] = coeffs
                        else:
                            kwargs["coeffs"] = args[i : i + num_coeffs]
                        i += num_coeffs
                    elif (
                        model
                        in ["MULTI_GAUSSIAN_KAPPA", "MULTI_GAUSSIAN_KAPPA_ELLIPSE"]
                        and name == "amp"
                    ):
                        if "sigma" in kwargs_fixed:
                            num_param = len(kwargs_fixed["sigma"])
                        else:
                            num_param = len(kwargs["sigma"])
                        kwargs["amp"] = args[i : i + num_param]
                        i += num_param
                    elif (
                        model
                        in ["MULTI_GAUSSIAN_KAPPA", "MULTI_GAUSSIAN_KAPPA_ELLIPSE"]
                        and name == "sigma"
                    ):
                        raise ValueError("%s must have fixed 'sigma' list!" % model)
                    elif model in ["INTERPOL", "INTERPOL_SCALED"] and name in [
                        "f_",
                        "f_xx",
                        "f_xy",
                        "f_yy",
                    ]:
                        pass
                    else:
                        kwargs[name] = args[i]
                        i += 1
                else:
                    kwargs[name] = kwargs_fixed[name]

                if name in kwargs_logsampling and name not in kwargs_fixed:
                    kwargs[name] = 10 ** (kwargs[name])

            kwargs_list.append(kwargs)

        return kwargs_list, i

    def set_params(self, kwargs_list):
        """

        :param kwargs_list: keyword argument list of lens model components
        :return: tuple of arguments (floats) that are being sampled
        """

        args = []
        for k, model in enumerate(self.model_list):
            kwargs = kwargs_list[k]
            kwargs_fixed = self.kwargs_fixed[k]
            kwargs_logsampling = self.kwargs_logsampling[k]

            param_names = self._param_name_list[k]
            for name in param_names:
                if name not in kwargs_fixed:
                    if (
                        model in ["SHAPELETS_POLAR", "SHAPELETS_CART"]
                        and name == "coeffs"
                    ):
                        coeffs = kwargs["coeffs"]
                        if self._solver_type == "SHAPELETS" and k == 0:
                            if self._num_images == 4:
                                coeffs = coeffs[6:]
                            elif self._num_images == 2:
                                coeffs = coeffs[3:]
                        args += list(coeffs)
                    elif (
                        model
                        in ["MULTI_GAUSSIAN_KAPPA", "MULTI_GAUSSIAN_KAPPA_ELLIPSE"]
                        and name == "amp"
                    ):
                        amp = kwargs["amp"]
                        args += list(amp)
                    elif (
                        model
                        in ["MULTI_GAUSSIAN_KAPPA", "MULTI_GAUSSIAN_KAPPA_ELLIPSE"]
                        and name == "sigma"
                    ):
                        raise ValueError("%s must have fixed 'sigma' list!" % model)
                    elif model in ["INTERPOL", "INTERPOL_SCALED"] and name in [
                        "f_",
                        "f_xx",
                        "f_xy",
                        "f_yy",
                    ]:
                        pass
                    # elif self._solver_type == 'PROFILE_SHEAR' and k == 1:
                    #    if name == 'e1':
                    #        _, gamma_ext = param_util.ellipticity2phi_gamma(kwargs['e1'], kwargs['e2'])
                    #        args.append(gamma_ext)
                    #    else:
                    #        pass
                    else:
                        if name in kwargs_logsampling:
                            args.append(np.log10(kwargs[name]))
                        else:
                            args.append(kwargs[name])

        return args

    def num_param(self):
        """

        :return: integer, number of free parameters being sampled from the lens model components
        """
        num = 0
        list = []
        type = "lens"
        for k, model in enumerate(self.model_list):
            kwargs_fixed = self.kwargs_fixed[k]
            param_names = self._param_name_list[k]
            for name in param_names:
                if name not in kwargs_fixed:
                    if (
                        model in ["SHAPELETS_POLAR", "SHAPELETS_CART"]
                        and name == "coeffs"
                    ):
                        num_coeffs = self._num_shapelet_lens
                        if self._solver_type == "SHAPELETS" and k == 0:
                            if self._num_images == 4:
                                num_coeffs -= 6
                            elif self._num_images == 2:
                                num_coeffs -= 3
                        num += num_coeffs
                        list += [str(name + "_" + type + str(k))] * num_coeffs
                    elif (
                        model
                        in ["MULTI_GAUSSIAN_KAPPA", "MULTI_GAUSSIAN_KAPPA_ELLIPSE"]
                        and name == "amp"
                    ):
                        num_param = len(kwargs_fixed["sigma"])
                        num += num_param
                        for i in range(num_param):
                            list.append(str(name + "_" + type + str(k)))
                    elif (
                        model
                        in ["MULTI_GAUSSIAN_KAPPA", "MULTI_GAUSSIAN_KAPPA_ELLIPSE"]
                        and name == "sigma"
                    ):
                        raise ValueError(
                            "'sigma' must be a fixed keyword argument for MULTI_GAUSSIAN"
                        )
                    elif model in ["INTERPOL", "INTERPOL_SCALED"] and name in [
                        "f_",
                        "f_xx",
                        "f_xy",
                        "f_yy",
                    ]:
                        pass
                    else:
                        num += 1
                        list.append(str(name + "_" + type + str(k)))
        return num, list



import lenstronomy.Util.util as util
import lenstronomy.Util.image_util as image_util

import numpy as np

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def data_configure_simple(
    numPix,
    deltaPix,
    exposure_time=None,
    background_rms=None,
    center_ra=0,
    center_dec=0,
    inverse=False,
):
    """Configures the data keyword arguments with a coordinate grid centered at zero.

    :param numPix: number of pixel (numPix x numPix)
    :param deltaPix: pixel size (in angular units)
    :param exposure_time: exposure time
    :param background_rms: background noise (Gaussian sigma)
    :param center_ra: RA at the center of the image
    :param center_dec: DEC at the center of the image
    :param inverse: if True, coordinate system is ra to the left, if False, to the right
    :return: keyword arguments that can be used to construct a Data() class instance of
        lenstronomy
    """

    # 1d list of coordinates (x,y) of a numPix x numPix square grid, centered to zero
    (
        x_grid,
        y_grid,
        ra_at_xy_0,
        dec_at_xy_0,
        x_at_radec_0,
        y_at_radec_0,
        Mpix2coord,
        Mcoord2pix,
    ) = util.make_grid_with_coordtransform(
        numPix=numPix,
        deltapix=deltaPix,
        center_ra=center_ra,
        center_dec=center_dec,
        subgrid_res=1,
        inverse=inverse,
    )
    # mask (1= model this pixel, 0= leave blanck)
    # exposure_map = np.ones((numPix, numPix)) * exposure_time  # individual exposure time/weight per pixel

    kwargs_data = {
        "background_rms": background_rms,
        "exposure_time": exposure_time,
        "ra_at_xy_0": ra_at_xy_0,
        "dec_at_xy_0": dec_at_xy_0,
        "transform_pix2angle": Mpix2coord,
        "image_data": np.zeros((numPix, numPix)),
    }
    return kwargs_data


@export
def simulate_simple(
    image_model_class,
    kwargs_lens=None,
    kwargs_source=None,
    kwargs_lens_light=None,
    kwargs_ps=None,
    no_noise=False,
    source_add=True,
    lens_light_add=True,
    point_source_add=True,
):
    """

    :param image_model_class:
    :param kwargs_lens:
    :param kwargs_source:
    :param kwargs_lens_light:
    :param kwargs_ps:
    :param no_noise:
    :param source_add:
    :param lens_light_add:
    :param point_source_add:
    :return:
    """

    image = image_model_class.image(
        kwargs_lens,
        kwargs_source,
        kwargs_lens_light,
        kwargs_ps,
        source_add=source_add,
        lens_light_add=lens_light_add,
        point_source_add=point_source_add,
    )
    # add noise
    if no_noise:
        return image
    else:
        poisson = image_util.add_poisson(
            image, exp_time=image_model_class.Data.exposure_map
        )
        bkg = image_util.add_background(
            image, sigma_bkd=image_model_class.Data.background_rms
        )
        return image + bkg + poisson



from lenstronomy.SimulationAPI.observation_api import SingleBand
from lenstronomy.Data.imaging_data import ImageData
import lenstronomy.Util.util as util
import numpy as np

__all__ = ["DataAPI"]


class DataAPI(SingleBand):
    """This class is a wrapper of the general description of data in SingleBand() to
    translate those quantities into configurations in the core lenstronomy Data modules
    to simulate images according to those quantities.

    This class is meant to be an example of a wrapper. More possibilities in terms of
    PSF and data type options are available. Have a look in the specific modules if you
    are interested in.
    """

    def __init__(self, numpix, kwargs_pixel_grid=None, **kwargs_single_band):
        """

        :param numpix: number of pixels per axis in the simulation to be modelled
        :param kwargs_pixel_grid: if None, uses default pixel grid option
            if defined, must contain keyword arguments PixelGrid() class
        :param kwargs_single_band: keyword arguments used to create instance of SingleBand class
        """
        self.numpix = numpix
        if kwargs_pixel_grid is not None:
            required_keys = ["ra_at_xy_0", "dec_at_xy_0", "transform_pix2angle"]
            if not all(k in kwargs_pixel_grid for k in required_keys):
                raise ValueError(
                    "Missing 1 or more required" + "kwargs_pixel_grid parameters"
                )
        self._kwargs_pixel_grid = kwargs_pixel_grid
        SingleBand.__init__(self, **kwargs_single_band)

    @property
    def data_class(self):
        """Creates a Data() instance of lenstronomy based on knowledge of the
        observation.

        :return: instance of Data() class
        """
        data_class = ImageData(**self.kwargs_data)
        return data_class

    @property
    def kwargs_data(self):
        """

        :return: keyword arguments for ImageData class instance
        """
        # default pixel grid
        if self._kwargs_pixel_grid is None:
            (
                _,
                _,
                ra_at_xy_0,
                dec_at_xy_0,
                _,
                _,
                transform_pix2angle,
                _,
            ) = util.make_grid_with_coordtransform(
                numPix=self.numpix,
                deltapix=self.pixel_scale,
                subgrid_res=1,
                left_lower=False,
                inverse=False,
            )
        # user defined pixel grid
        else:
            ra_at_xy_0 = self._kwargs_pixel_grid["ra_at_xy_0"]
            dec_at_xy_0 = self._kwargs_pixel_grid["dec_at_xy_0"]
            transform_pix2angle = self._kwargs_pixel_grid["transform_pix2angle"]
        # CCD gain corrected exposure time to allow a direct Poisson estimates based on IID counts
        scaled_exposure_time = self.flux_iid(1)
        kwargs_data = {
            "image_data": np.zeros((self.numpix, self.numpix)),
            "ra_at_xy_0": ra_at_xy_0,
            "dec_at_xy_0": dec_at_xy_0,
            "transform_pix2angle": transform_pix2angle,
            "background_rms": self.background_noise,
            "exposure_time": scaled_exposure_time,
        }
        return kwargs_data



from lenstronomy.LensModel.Profiles.hernquist import Hernquist
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
import numpy as np

__all__ = ["Hernquist_Ellipse"]


class Hernquist_Ellipse(LensProfileBase):
    """This class contains functions for the elliptical Hernquist profile.

    Ellipticity is defined in the potential.
    """

    param_names = ["sigma0", "Rs", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "sigma0": 0,
        "Rs": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "sigma0": 100,
        "Rs": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.spherical = Hernquist()
        self._diff = 0.00000001
        super(Hernquist_Ellipse, self).__init__()

    def function(self, x, y, sigma0, Rs, e1, e2, center_x=0, center_y=0):
        """Returns double integral of NFW profile."""
        x_, y_ = param_util.transform_e1e2_square_average(
            x, y, e1, e2, center_x, center_y
        )
        f_ = self.spherical.function(x_, y_, sigma0, Rs)
        return f_

    def derivatives(self, x, y, sigma0, Rs, e1, e2, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function (integral of NFW)"""
        x_, y_ = param_util.transform_e1e2_square_average(
            x, y, e1, e2, center_x, center_y
        )
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        cos_phi = np.cos(phi_G)
        sin_phi = np.sin(phi_G)
        e = param_util.q2e(q)

        f_x_prim, f_y_prim = self.spherical.derivatives(x_, y_, sigma0, Rs)
        f_x_prim *= np.sqrt(1 - e)
        f_y_prim *= np.sqrt(1 + e)
        f_x = cos_phi * f_x_prim - sin_phi * f_y_prim
        f_y = sin_phi * f_x_prim + cos_phi * f_y_prim
        return f_x, f_y

    def hessian(self, x, y, sigma0, Rs, e1, e2, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx,
        d^f/dy^2."""
        alpha_ra, alpha_dec = self.derivatives(
            x, y, sigma0, Rs, e1, e2, center_x, center_y
        )
        diff = self._diff
        alpha_ra_dx, alpha_dec_dx = self.derivatives(
            x + diff, y, sigma0, Rs, e1, e2, center_x, center_y
        )
        alpha_ra_dy, alpha_dec_dy = self.derivatives(
            x, y + diff, sigma0, Rs, e1, e2, center_x, center_y
        )

        f_xx = (alpha_ra_dx - alpha_ra) / diff
        f_xy = (alpha_ra_dy - alpha_ra) / diff
        f_yx = (alpha_dec_dx - alpha_dec) / diff
        f_yy = (alpha_dec_dy - alpha_dec) / diff
        return f_xx, f_xy, f_yx, f_yy

    def density(self, r, rho0, Rs, e1=0, e2=0):
        """Computes the 3-d density.

        :param r: 3-d radius
        :param rho0: density normalization
        :param Rs: Hernquist radius
        :return: density at radius r
        """
        return self.spherical.density(r, rho0, Rs)

    def density_lens(self, r, sigma0, Rs, e1=0, e2=0):
        """Density as a function of 3d radius in lensing parameters This function
        converts the lensing definition sigma0 into the 3d density.

        :param r: 3d radius
        :param sigma0: rho0 * Rs (units of projected density)
        :param Rs: Hernquist radius
        :return: enclosed mass in 3d
        """
        return self.spherical.density_lens(r, sigma0, Rs)

    def density_2d(self, x, y, rho0, Rs, e1=0, e2=0, center_x=0, center_y=0):
        """Projected density along the line of sight at coordinate (x, y)

        :param x: x-coordinate
        :param y: y-coordinate
        :param rho0: density normalization
        :param Rs: Hernquist radius
        :param center_x: x-center of the profile
        :param center_y: y-center of the profile
        :return: projected density
        """
        return self.spherical.density_2d(x, y, rho0, Rs, center_x, center_y)

    def mass_2d_lens(self, r, sigma0, Rs, e1=0, e2=0):
        """Mass enclosed projected 2d sphere of radius r Same as mass_2d but with input
        normalization in units of projected density.

        :param r: projected radius
        :param sigma0: rho0 * Rs (units of projected density)
        :param Rs: Hernquist radius
        :return: mass enclosed 2d projected radius
        """
        return self.spherical.mass_2d_lens(r, sigma0, Rs)

    def mass_2d(self, r, rho0, Rs, e1=0, e2=0):
        """Mass enclosed projected 2d sphere of radius r.

        :param r: projected radius
        :param rho0: density normalization
        :param Rs: Hernquist radius
        :return: mass enclosed 2d projected radius
        """
        return self.spherical.mass_2d(r, rho0, Rs)

    def mass_3d(self, r, rho0, Rs, e1=0, e2=0):
        """Mass enclosed a 3d sphere or radius r.

        :param r: 3-d radius within the mass is integrated (same distance units as
            density definition)
        :param rho0: density normalization
        :param Rs: Hernquist radius
        :return: enclosed mass
        """
        return self.spherical.mass_3d(r, rho0, Rs)

    def mass_3d_lens(self, r, sigma0, Rs, e1=0, e2=0):
        """Mass enclosed a 3d sphere or radius r in lensing parameterization.

        :param r: 3-d radius within the mass is integrated (same distance units as
            density definition)
        :param sigma0: rho0 * Rs (units of projected density)
        :param Rs: Hernquist radius
        :return: enclosed mass
        """
        return self.spherical.mass_3d_lens(r, sigma0, Rs)



# Conditions for the ZTF survey

import lenstronomy.Util.util as util

__all__ = ["ZTF"]
"""
Sources
  https://iopscience.iop.org/article/10.1088/1538-3873/aaecbe/pdf
    Table 1
      - seeing
      - pixel_scale
      - gain
      - read_noise

  http://web.ipac.caltech.edu/staff/fmasci/ztf/ztf_pipelines_deliverables.pdf
    Section 10.1.2
      - magnitude_zero_point

  https://www.ztf.caltech.edu/page/dr3
    Section 10
      - exposure_time (assuming public survey only)

  https://iopscience.iop.org/article/10.1088/1538-3873/aae8ac/pdf
    Section 3.7
      - num_exposures 
"""

g_band_obs = {
    "exposure_time": 30.0,
    "sky_brightness": 22.01,
    "magnitude_zero_point": 26.325,
    "num_exposures": 40,
    "seeing": 2.1,
    "psf_type": "GAUSSIAN",
}

r_band_obs = {
    "exposure_time": 30.0,
    "sky_brightness": 21.15,
    "magnitude_zero_point": 26.275,
    "num_exposures": 40,
    "seeing": 2.0,
    "psf_type": "GAUSSIAN",
}

i_band_obs = {
    "exposure_time": 30.0,
    "sky_brightness": 19.89,
    "magnitude_zero_point": 25.660,
    "num_exposures": 40,
    "seeing": 2.1,
    "psf_type": "GAUSSIAN",
}

# - keyword exposure_time: exposure time per image (in seconds)
# - keyword sky_brightness: sky brightness (in magnitude per square arcseconds in units of electrons)
# - keyword magnitude_zero_point: magnitude in which 1 count (e-) per second per arcsecond square is registered
# - keyword num_exposures: number of exposures that are combined (depends on coadd_years)
# - keyword seeing: Full-Width-at-Half-Maximum (FWHM) of PSF
# - keyword psf_type: string, type of PSF ('GAUSSIAN' supported)


class ZTF(object):
    """Class contains ZTF instrument and observation configurations."""

    def __init__(self, band="g", psf_type="GAUSSIAN", coadd_years=3):
        """
        :param band: string, 'g', 'r', or 'i', supported. Determines obs dictionary.
        :param psf_type: string, type of PSF ('GAUSSIAN' supported).
        :param coadd_years: int, number of years corresponding to num_exposures in obs dict. Currently supported: 1-3.
        """
        if band == "g":
            self.obs = g_band_obs
        elif band == "r":
            self.obs = r_band_obs
        elif band == "i":
            self.obs = i_band_obs
        else:
            raise ValueError("band %s not supported! Choose 'g', 'r', or 'i'." % band)

        if psf_type != "GAUSSIAN":
            raise ValueError("psf_type %s not supported!" % psf_type)

        if coadd_years > 3 or coadd_years < 1:
            raise ValueError(
                " %s coadd_years not supported! Choose an integer between 1 and 3."
                % coadd_years
            )
        elif coadd_years != 3:
            self.obs["num_exposures"] = (coadd_years * 40) // 3

        self.camera = {"read_noise": 10.3, "pixel_scale": 1.01, "ccd_gain": 5.8}

        # - keyword read_noise: std of noise generated by read-out (in units of electrons)
        # - keyword pixel_scale: scale (in arcseconds) of pixels
        # - keyword ccd_gain: electrons/ADU (analog-to-digital unit)

    def kwargs_single_band(self):
        """
        :return: merged kwargs from camera and obs dicts
        """
        kwargs = util.merge_dicts(self.camera, self.obs)
        return kwargs



__author__ = "mgomer"

from lenstronomy.LensModel.Profiles.synthesis import SynthesisProfile
from lenstronomy.Analysis.lens_profile import LensProfileAnalysis
from lenstronomy.LensModel.lens_model import LensModel

import numpy as np
import numpy.testing as npt
import pytest


class TestSynthesis(object):
    """Tests the synthesis model's ability to approximate several profiles."""

    def setup_method(self):
        self.lin_fit_hyperparams = {"num_r_evals": 100}
        self.s_list = np.logspace(-6.0, 3.0, 30)
        self.x_test = np.linspace(0.01, 2, 10)
        self.y_test = np.zeros_like(self.x_test)

    def test_CSE_components(self):
        # test potential, deflection, and kappa using CSE components for a few profiles
        kwargs_list = []
        center_x = 0
        center_y = 0
        for s in self.s_list:
            kwargs_list.append(
                {
                    "a": 1,
                    "s": s,
                    "e1": 0,
                    "e2": 0,
                    "center_x": center_x,
                    "center_y": center_y,
                }
            )

        # test nfw
        kwargs_nfw = [
            {"Rs": 1.5, "alpha_Rs": 1, "center_x": center_x, "center_y": center_y}
        ]
        kwargs_synthesis = {
            "target_lens_model": "NFW",
            "component_lens_model": "CSE",
            "kwargs_list": kwargs_list,
            "lin_fit_hyperparams": self.lin_fit_hyperparams,
        }
        lensmodel_synth = LensModel(["SYNTHESIS"], kwargs_synthesis=kwargs_synthesis)
        lensmodel_nfw = LensModel(["NFW"])
        self.compare_synth(lensmodel_synth, lensmodel_nfw, kwargs_nfw)
        # test sersic
        kwargs_sersic = [{"k_eff": 3, "R_sersic": 1.5, "n_sersic": 3.5}]
        kwargs_synthesis = {
            "target_lens_model": "SERSIC",
            "component_lens_model": "CSE",
            "kwargs_list": kwargs_list,
            "lin_fit_hyperparams": self.lin_fit_hyperparams,
        }
        lensmodel_synth = LensModel(["SYNTHESIS"], kwargs_synthesis=kwargs_synthesis)
        lensmodel_sersic = LensModel(["SERSIC"])
        self.compare_synth(lensmodel_synth, lensmodel_sersic, kwargs_sersic)
        # test hernquist
        kwargs_hernquist = [
            {"sigma0": 10, "Rs": 1.5, "center_x": center_x, "center_y": center_y}
        ]
        kwargs_synthesis = {
            "target_lens_model": "HERNQUIST",
            "component_lens_model": "CSE",
            "kwargs_list": kwargs_list,
            "lin_fit_hyperparams": self.lin_fit_hyperparams,
        }
        lensmodel_synth = LensModel(["SYNTHESIS"], kwargs_synthesis=kwargs_synthesis)
        lensmodel_hernquist = LensModel(["HERNQUIST"])
        self.compare_synth(lensmodel_synth, lensmodel_hernquist, kwargs_hernquist)

    def test_gaussian_components(self):
        # test potential, deflection, and kappa using Gaussian components for a few profiles
        kwargs_list = []
        center_x = 0
        center_y = 0
        for s in self.s_list:
            kwargs_list.append(
                {"amp": 1, "sigma": s, "e1": 0, "e2": 0, "center_x": 0, "center_y": 0}
            )
        # test nfw
        kwargs_nfw = [
            {"Rs": 1.5, "alpha_Rs": 1, "center_x": center_x, "center_y": center_y}
        ]
        kwargs_synthesis = {
            "target_lens_model": "NFW",
            "component_lens_model": "GAUSSIAN_ELLIPSE_KAPPA",
            "kwargs_list": kwargs_list,
            "lin_fit_hyperparams": self.lin_fit_hyperparams,
        }
        lensmodel_synth = LensModel(["SYNTHESIS"], kwargs_synthesis=kwargs_synthesis)
        lensmodel_nfw = LensModel(["NFW"])
        self.compare_synth(lensmodel_synth, lensmodel_nfw, kwargs_nfw)
        # test sersic
        kwargs_sersic = [{"k_eff": 3, "R_sersic": 1.5, "n_sersic": 3.5}]
        kwargs_synthesis = {
            "target_lens_model": "SERSIC",
            "component_lens_model": "GAUSSIAN_ELLIPSE_KAPPA",
            "kwargs_list": kwargs_list,
            "lin_fit_hyperparams": self.lin_fit_hyperparams,
        }
        lensmodel_synth = LensModel(["SYNTHESIS"], kwargs_synthesis=kwargs_synthesis)
        lensmodel_sersic = LensModel(["SERSIC"])
        self.compare_synth(lensmodel_synth, lensmodel_sersic, kwargs_sersic)
        # test hernquist
        kwargs_hernquist = [
            {"sigma0": 10, "Rs": 1.5, "center_x": center_x, "center_y": center_y}
        ]
        kwargs_synthesis = {
            "target_lens_model": "HERNQUIST",
            "component_lens_model": "GAUSSIAN_ELLIPSE_KAPPA",
            "kwargs_list": kwargs_list,
            "lin_fit_hyperparams": self.lin_fit_hyperparams,
        }
        lensmodel_synth = LensModel(["SYNTHESIS"], kwargs_synthesis=kwargs_synthesis)
        lensmodel_hernquist = LensModel(["HERNQUIST"])
        self.compare_synth(lensmodel_synth, lensmodel_hernquist, kwargs_hernquist)

    def compare_synth(self, lensmodel_synth, lensmodel_target, kwargs_target):
        # check potentials, deflection, kappa
        synth_pot = lensmodel_synth.potential(self.x_test, self.y_test, kwargs_target)
        target_pot = lensmodel_target.potential(self.x_test, self.y_test, kwargs_target)
        synth_defl = lensmodel_synth.alpha(self.x_test, self.y_test, kwargs_target)[0]
        target_defl = lensmodel_target.alpha(self.x_test, self.y_test, kwargs_target)[0]
        synth_kappa = lensmodel_synth.kappa(self.x_test, self.y_test, kwargs_target)
        target_kappa = lensmodel_target.kappa(self.x_test, self.y_test, kwargs_target)

        npt.assert_allclose(synth_pot, target_pot, rtol=1e-2)  # potentials within 1%
        npt.assert_allclose(synth_defl, target_defl, rtol=1e-2)  # deflections within 1%
        npt.assert_allclose(synth_kappa, target_kappa, rtol=1e-2)  # kappa within 1%

    def test_ellipticity_and_centers(self):
        # check that even with ellipticity and offset center, same radial profile
        r_test = self.x_test
        kwargs_list = []
        center_x = 1
        center_y = -1
        for s in self.s_list:
            kwargs_list.append(
                {
                    "a": 1,
                    "s": s,
                    "e1": 0.3,
                    "e2": 0,
                    "center_x": center_x,
                    "center_y": center_y,
                }
            )

        # test nfw from CSEs
        kwargs_nfw = [
            {"Rs": 1.5, "alpha_Rs": 1, "center_x": center_x, "center_y": center_y}
        ]
        kwargs_synthesis = {
            "target_lens_model": "NFW",
            "component_lens_model": "CSE",
            "kwargs_list": kwargs_list,
            "lin_fit_hyperparams": self.lin_fit_hyperparams,
        }
        lensmodel_synth = LensModel(["SYNTHESIS"], kwargs_synthesis=kwargs_synthesis)
        lensmodel_nfw = LensModel(["NFW"])
        LensAn_synth = LensProfileAnalysis(lensmodel_synth)
        synth_avg_kappa = LensAn_synth.radial_lens_profile(r_test, kwargs_nfw)
        LensAn_nfw = LensProfileAnalysis(lensmodel_nfw)
        nfw_avg_kappa = LensAn_nfw.radial_lens_profile(r_test, kwargs_nfw)
        npt.assert_allclose(nfw_avg_kappa, synth_avg_kappa, rtol=5e-2)

    def test_set_static_method(self):
        # test that when set_static is used the result doesn't load the kwargs anymore, instead using the saved weights
        kwargs_list = []
        center_x = 0
        center_y = 0
        for s in self.s_list:
            kwargs_list.append(
                {
                    "a": 1,
                    "s": s,
                    "e1": 0,
                    "e2": 0,
                    "center_x": center_x,
                    "center_y": center_y,
                }
            )

        # test nfw
        kwargs_nfw = [
            {"Rs": 1.5, "alpha_Rs": 1, "center_x": center_x, "center_y": center_y}
        ]
        kwargs_synthesis = {
            "target_lens_model": "NFW",
            "component_lens_model": "CSE",
            "kwargs_list": kwargs_list,
            "lin_fit_hyperparams": self.lin_fit_hyperparams,
        }
        synth_prof = SynthesisProfile(**kwargs_synthesis)
        synth_func = synth_prof.function(self.x_test, self.y_test, **kwargs_nfw[0])
        linear_weights = synth_prof.linear_weight_mle_fit(kwargs_nfw, kwargs_list)
        synth_prof.set_static(linear_weights)
        kwargs_nfw[0]["Rs"] = 10  # change kwargs, doesn't affect static weights
        changed_synth_func = synth_prof.function(
            self.x_test, self.y_test, **kwargs_nfw[0]
        )
        npt.assert_array_equal(synth_func, changed_synth_func)



__author__ = "sibirrer"

import pytest
import numpy as np
import numpy.testing as npt
import unittest
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.ImSim.image_model import ImageModel
import lenstronomy.Util.param_util as param_util
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.Plots.model_plot import check_solver_error
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
from lenstronomy.Analysis.image_reconstruction import MultiBandImageReconstruction


class TestMultiBandImageReconstruction(object):
    """Test the fitting sequences."""

    def setup_method(self):
        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 10  # cutout pixel size
        deltaPix = 0.5  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification

        self.kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        data_class = ImageData(**self.kwargs_data)
        self.kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": deltaPix,
        }
        psf_class = PSF(**self.kwargs_psf)

        # 'EXERNAL_SHEAR': external shear
        kwargs_shear = {
            "gamma1": 0.01,
            "gamma2": 0.01,
        }  # gamma_ext: shear strength, psi_ext: shear angel (in radian)
        e1, e2 = param_util.phi_q2_ellipticity(0.2, 0.8)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": 1.8,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_model_list = ["SPEP", "SHEAR"]
        self.kwargs_lens = [kwargs_spemd, kwargs_shear]
        lens_model_class = LensModel(lens_model_list=lens_model_list)
        self.LensModel = lens_model_class
        # list of light profiles (for lens and source)
        # 'SERSIC': spherical Sersic profile
        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        phi, q = 0.2, 0.9
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 7,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_light_model_list = ["SERSIC"]
        self.kwargs_lens_light = [kwargs_sersic]
        lens_light_model_class = LightModel(light_model_list=lens_light_model_list)
        source_model_list = ["SERSIC_ELLIPSE"]
        self.kwargs_source = [kwargs_sersic_ellipse]
        source_model_class = LightModel(light_model_list=source_model_list)
        self.kwargs_ps = [
            {"ra_source": 0.0, "dec_source": 0.0, "source_amp": 1.0}
        ]  # quasar point source position in the source plane and intrinsic brightness
        point_source_list = ["SOURCE_POSITION"]
        point_source_class = PointSource(
            point_source_type_list=point_source_list, fixed_magnification_list=[True]
        )
        kwargs_numerics = {"supersampling_factor": 1}
        imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )
        image_sim = sim_util.simulate_simple(
            imageModel,
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
            no_noise=True,
        )

        data_class.update_data(image_sim)
        self.kwargs_data["image_data"] = image_sim
        self.kwargs_model = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": source_model_list,
            "lens_light_model_list": lens_light_model_list,
            "point_source_model_list": point_source_list,
            "fixed_magnification_list": [False],
        }
        self.kwargs_numerics = kwargs_numerics
        self.data_class = ImageData(**self.kwargs_data)
        self.kwargs_params = {
            "kwargs_lens": self.kwargs_lens,
            "kwargs_source": self.kwargs_source,
            "kwargs_lens_light": self.kwargs_lens_light,
            "kwargs_ps": self.kwargs_ps,
        }

    def test_band_setup(self):
        multi_band_list = [[self.kwargs_data, self.kwargs_psf, self.kwargs_numerics]]
        multi_band = MultiBandImageReconstruction(
            multi_band_list,
            self.kwargs_model,
            self.kwargs_params,
            multi_band_type="single-band",
        )

        multi_band = MultiBandImageReconstruction(
            multi_band_list,
            self.kwargs_model,
            self.kwargs_params,
            multi_band_type="joint-linear",
        )

        image_model, kwargs_params = multi_band.band_setup(band_index=0)
        model = image_model.image(**kwargs_params)
        npt.assert_almost_equal(model, self.kwargs_data["image_data"], decimal=5)
        npt.assert_almost_equal(model, multi_band.model_band_list[0].model, decimal=5)
        npt.assert_almost_equal(
            multi_band.model_band_list[0].norm_residuals, 0, decimal=5
        )

    def test_bands_compute(self):
        multi_band_list = [
            [self.kwargs_data, self.kwargs_psf, self.kwargs_numerics]
        ] * 2
        multi_band = MultiBandImageReconstruction(
            multi_band_list,
            self.kwargs_model,
            self.kwargs_params,
            kwargs_likelihood={"bands_compute": [False, True]},
            multi_band_type="multi-linear",
        )
        image_model, kwargs_params = multi_band.band_setup(band_index=1)
        model = image_model.image(**kwargs_params)
        npt.assert_almost_equal(model, self.kwargs_data["image_data"], decimal=5)

    def test_not_verbose(self):
        multi_band_list = [[self.kwargs_data, self.kwargs_psf, self.kwargs_numerics]]
        multi_band = MultiBandImageReconstruction(
            multi_band_list, self.kwargs_model, self.kwargs_params, verbose=False
        )

    def test_check_solver_error(self):
        bool = check_solver_error(image=np.array([0, 0]))
        assert bool

        bool = check_solver_error(image=np.array([0, 0.1]))
        assert bool == 0


class TestRaises(unittest.TestCase):
    def test_no_band(self):
        """Test raise statements if band is not evaluated."""
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 10  # cutout pixel size
        deltaPix = 0.5  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        kwargs_data["image_data"] = np.ones((numPix, numPix))
        kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": deltaPix,
        }
        kwargs_numerics = {}
        multi_band_list = [[kwargs_data, kwargs_psf, kwargs_numerics]]
        multi_band = MultiBandImageReconstruction(
            multi_band_list,
            {},
            {},
            multi_band_type="single-band",
            kwargs_likelihood={"bands_compute": [False]},
        )
        with self.assertRaises(ValueError):
            multi_band.band_setup(band_index=0)


if __name__ == "__main__":
    pytest.main()



import numpy as np
import numpy.testing as npt
import pytest
import unittest

from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.LightModel.Profiles.gaussian import Gaussian
from lenstronomy.LightModel.Profiles.starlets import SLIT_Starlets
from lenstronomy.Util import util


_force_no_pysap = (
    True  # if issues on Travis-CI to install pysap, force use python-only functions
)


class TestSLITStarlets(object):
    """Class to test SLIT_Starlets light profile."""

    def setup_method(self):
        # different versions of Starlet transforms
        self.starlets = SLIT_Starlets(
            fast_inverse=False, second_gen=False, force_no_pysap=_force_no_pysap
        )
        self.starlets_fast = SLIT_Starlets(
            fast_inverse=True, second_gen=False, force_no_pysap=_force_no_pysap
        )
        self.starlets_2nd = SLIT_Starlets(
            second_gen=True, force_no_pysap=_force_no_pysap
        )

        # define a test image with gaussian components
        self.num_pix = 50
        self.n_scales = 3
        self.n_pixels = self.num_pix**2
        self.x, self.y = util.make_grid(self.num_pix, 1)

        # build a non-trivial positive image from sum of gaussians
        gaussian = Gaussian()
        gaussian1 = gaussian.function(
            self.x, self.y, amp=100, sigma=1, center_x=-7, center_y=-7
        )
        gaussian2 = gaussian.function(
            self.x, self.y, amp=500, sigma=3, center_x=-3, center_y=-3
        )
        gaussian3 = gaussian.function(
            self.x, self.y, amp=2000, sigma=5, center_x=+5, center_y=+5
        )
        self.test_image = util.array2image(gaussian1 + gaussian2 + gaussian3)

        self.test_coeffs = np.zeros((self.n_scales, self.num_pix, self.num_pix))
        self.test_coeffs[0, :, :] = util.array2image(gaussian1)
        self.test_coeffs[1, :, :] = util.array2image(gaussian2)
        self.test_coeffs[2, :, :] = util.array2image(gaussian3)

    def test_reconstructions_2d(self):
        """

        :return:
        """
        # PySAP requires call to decomposition once before anything else
        self.starlets.decomposition_2d(self.test_image, self.n_scales)
        self.starlets_fast.decomposition_2d(self.test_image, self.n_scales)
        self.starlets_2nd.decomposition_2d(self.test_image, self.n_scales)

        image = self.starlets.function_2d(
            coeffs=self.test_coeffs, n_scales=self.n_scales, n_pixels=self.n_pixels
        )
        image_fast = self.starlets_fast.function_2d(
            coeffs=self.test_coeffs, n_scales=self.n_scales, n_pixels=self.n_pixels
        )
        assert image.shape == (self.num_pix, self.num_pix)
        assert image_fast.shape == (self.num_pix, self.num_pix)

        image_2nd = self.starlets_2nd.function_2d(
            coeffs=self.test_coeffs, n_scales=self.n_scales, n_pixels=self.n_pixels
        )
        assert image_2nd.shape == (self.num_pix, self.num_pix)
        assert np.all(image_2nd >= 0)

    def test_decompositions_2d(self):
        """

        :return:
        """
        # test equality between fast and std transform (which are identical)
        coeffs = self.starlets.decomposition_2d(self.test_image, self.n_scales)
        coeffs_fast = self.starlets_fast.decomposition_2d(
            self.test_image, self.n_scales
        )
        assert coeffs.shape == (self.n_scales, self.num_pix, self.num_pix)
        assert coeffs_fast.shape == (self.n_scales, self.num_pix, self.num_pix)
        npt.assert_almost_equal(coeffs, coeffs_fast, decimal=3)

        # test non-negativity of second generation starlet transform
        coeffs_2nd = self.starlets_2nd.decomposition_2d(self.test_image, self.n_scales)
        assert coeffs_2nd.shape == (self.n_scales, self.num_pix, self.num_pix)

    def test_function(self):
        """

        :return:
        """
        # PySAP requires call to decomposition once before anything else
        self.starlets.decomposition(self.test_image, self.n_scales)
        self.starlets_fast.decomposition(self.test_image, self.n_scales)
        self.starlets_2nd.decomposition(self.test_image, self.n_scales)

        # test with a 1D input
        self.starlets.decomposition(util.image2array(self.test_image), self.n_scales)

        coeffs_1d = self.test_coeffs.reshape(self.n_scales * self.num_pix**2)

        image_1d = self.starlets.function(
            self.x,
            self.y,
            amp=coeffs_1d,
            n_scales=self.n_scales,
            n_pixels=self.n_pixels,
        )
        assert image_1d.shape == (self.num_pix**2,)
        image_1d_fast = self.starlets_fast.function(
            self.x,
            self.y,
            amp=coeffs_1d,
            n_scales=self.n_scales,
            n_pixels=self.n_pixels,
        )
        assert image_1d_fast.shape == (self.num_pix**2,)
        image_1d_2nd = self.starlets_2nd.function(
            self.x,
            self.y,
            amp=coeffs_1d,
            n_scales=self.n_scales,
            n_pixels=self.n_pixels,
        )
        assert image_1d_2nd.shape == (self.num_pix**2,)

    def test_identity_operations_fast(self):
        """Test the decomposition/reconstruction.

        :return:
        """
        coeffs = self.starlets_fast.decomposition_2d(self.test_image, self.n_scales)
        test_image_recon = self.starlets_fast.function_2d(
            coeffs=coeffs, n_scales=self.n_scales, n_pixels=self.n_pixels
        )
        npt.assert_almost_equal(self.test_image, test_image_recon, decimal=5)

    def test_identity_operations_2nd(self):
        """Test the decomposition/reconstruction.

        :return:
        """
        coeffs = self.starlets_2nd.decomposition_2d(self.test_image, self.n_scales)
        test_image_recon = self.starlets_2nd.function_2d(
            coeffs=coeffs, n_scales=self.n_scales, n_pixels=self.n_pixels
        )
        npt.assert_almost_equal(self.test_image, test_image_recon, decimal=5)

    def test_delete_cache(self):
        amp = self.test_coeffs.reshape(self.n_scales * self.num_pix**2)
        kwargs_starlets = dict(
            amp=amp,
            n_scales=self.n_scales,
            n_pixels=self.n_pixels,
            center_x=0,
            center_y=0,
            scale=1,
        )
        output = self.starlets_fast.function(self.x, self.y, **kwargs_starlets)
        assert hasattr(self.starlets_fast.interpol, "_image_interp")
        self.starlets_fast.delete_cache()
        assert not hasattr(self.starlets_fast.interpol, "_image_interp")

    def test_coeffs2pysap(self):
        n_scales = 3
        num_pix = 20
        coeffs = np.ones((n_scales, num_pix, num_pix))
        pysap_list = self.starlets._coeffs2pysap(coeffs)
        assert len(pysap_list) == n_scales
        for i in range(n_scales):
            assert pysap_list[i].shape == coeffs[i].shape

    def test_pysap2coeffs(self):
        n_scales = 3
        num_pix = 20
        pysap_list = n_scales * [np.ones((num_pix, num_pix))]
        coeffs = self.starlets._pysap2coeffs(pysap_list)
        assert coeffs.shape == (n_scales, num_pix, num_pix)
        for i in range(n_scales):
            assert pysap_list[i].shape == coeffs[i].shape


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            # try to set decomposition scale to higher than maximal value
            starlets = SLIT_Starlets(force_no_pysap=True)
            # define a test image with gaussian components
            num_pix = 50
            x, y = util.make_grid(num_pix, 1)
            # build a non-trivial positive image from sum of gaussians
            gaussian = Gaussian()
            gaussian1 = gaussian.function(
                x, y, amp=100, sigma=1, center_x=-7, center_y=-7
            )
            gaussian2 = gaussian.function(
                x, y, amp=500, sigma=3, center_x=-3, center_y=-3
            )
            gaussian3 = gaussian.function(
                x, y, amp=2000, sigma=5, center_x=+5, center_y=+5
            )
            test_image = util.array2image(gaussian1 + gaussian2 + gaussian3)
            n_scales = 100
            _ = starlets.decomposition_2d(test_image, n_scales)
        with self.assertRaises(ValueError):
            # try to set decomposition scale to negative value
            starlets = SLIT_Starlets(force_no_pysap=True)
            # define a test image with gaussian components
            num_pix = 50
            x, y = util.make_grid(num_pix, 1)
            # build a non-trivial positive image from sum of gaussians
            gaussian = Gaussian()
            gaussian1 = gaussian.function(
                x, y, amp=100, sigma=1, center_x=-7, center_y=-7
            )
            gaussian2 = gaussian.function(
                x, y, amp=500, sigma=3, center_x=-3, center_y=-3
            )
            gaussian3 = gaussian.function(
                x, y, amp=2000, sigma=5, center_x=+5, center_y=+5
            )
            test_image = util.array2image(gaussian1 + gaussian2 + gaussian3)
            n_scales = -1
            _ = starlets.decomposition_2d(test_image, n_scales)
        with self.assertRaises(ValueError):
            # function_split is not supported/defined for pixel-based profiles
            light_model = LightModel(["SLIT_STARLETS"])
            num_pix = 20
            x, y = util.make_grid(num_pix, 1)
            kwargs_list = [
                {
                    "amp": np.ones((3, num_pix, num_pix)),
                    "n_scales": 3,
                    "n_pixels": 20**2,
                    "center_x": 0,
                    "center_y": 0,
                    "scale": 1,
                }
            ]
            _ = light_model.functions_split(x, y, kwargs_list)
        with self.assertRaises(ValueError):
            # provided a wrong shape for starlet coefficients
            starlet_class = SLIT_Starlets()
            num_pix = 20
            x, y = util.make_grid(num_pix, 1)
            coeffs_wrong = np.ones((3, num_pix**2))
            kwargs_list = {
                "amp": coeffs_wrong,
                "n_scales": 3,
                "n_pixels": 20**2,
                "center_x": 0,
                "center_y": 0,
                "scale": 1,
            }
            _ = starlet_class.function(x, y, **kwargs_list)
            image_wrong = np.ones((1, num_pix, num_pix))
            _ = starlet_class.decomposition(image_wrong, 3)
        with self.assertRaises(ValueError):
            # provided a wrong shape for image to be decomposed
            starlet_class = SLIT_Starlets()
            num_pix = 20
            image_wrong = np.ones((2, num_pix, num_pix))
            _ = starlet_class.decomposition(image_wrong, 3)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"


from lenstronomy.LensModel.Profiles.convergence import Convergence

import numpy as np
import numpy.testing as npt
import pytest


class TestConvergence(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.profile = Convergence()
        self.kwargs_lens = {"kappa": 0.1}

    def test_function(self):
        x = np.array([1])
        y = np.array([0])
        values = self.profile.function(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(values[0], self.kwargs_lens["kappa"] / 2, decimal=5)
        x = np.array([0])
        y = np.array([0])
        values = self.profile.function(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(values[0], 0, decimal=5)

        x = np.array([2, 3, 4])
        y = np.array([1, 1, 1])
        values = self.profile.function(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(values[0], 0.25, decimal=5)
        npt.assert_almost_equal(values[1], 0.5, decimal=5)

    def test_derivatives(self):
        x = np.array([1])
        y = np.array([2])
        f_x, f_y = self.profile.derivatives(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(f_x[0], 0.1, decimal=5)
        npt.assert_almost_equal(f_y[0], 0.2, decimal=5)

        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        values = self.profile.derivatives(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(values[0][0], 0.1, decimal=5)
        npt.assert_almost_equal(values[1][0], 0.2, decimal=5)

    def test_hessian(self):
        x = np.array([1])
        y = np.array([2])
        f_xx, f_xy, f_yx, f_yy = self.profile.hessian(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(f_xx, 0.1, decimal=5)
        npt.assert_almost_equal(f_yy, 0.1, decimal=5)
        npt.assert_almost_equal(f_xy, 0, decimal=5)
        npt.assert_almost_equal(f_yx, 0, decimal=5)

        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        values = self.profile.hessian(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(values[0], 0.1, decimal=5)
        npt.assert_almost_equal(values[3], 0.1, decimal=5)
        npt.assert_almost_equal(values[1], 0, decimal=5)


if __name__ == "__main__":
    pytest.main()



__author__ = "mgomer"

from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
import numpy as np
import copy
from lenstronomy.Util import param_util
from lenstronomy.Util import util
from lenstronomy.LensModel.lens_model import LensModel


class SynthesisProfile(LensProfileBase):
    """A general class which describes a linear sum of many simple profiles to
    approximate a target profile.

    Example: Mimic an NFW profile with many CSE profiles. In this case, you could use LensModel(['SYNTHESIS'],kwargs_synthesis=kwargs_synthesis) with
    kwargs_synthesis={'target_lens_model': 'NFW',
                    'component_lens_model': 'CSE',
                   'kwargs_list': kwargs_list,
                   'lin_fit_hyperparams':{'lower_log_bound':-6, 'upper_log_bound':3, 'num_r_evals':100, 'sigma':0.01} (default values)
                   }
    where kwargs_list would be a list of input CSE kwargs (where the amplitude will be re-adjusted).
    """

    profile_name = "SYNTHESIS"

    def __init__(
        self, target_lens_model, component_lens_model, kwargs_list, lin_fit_hyperparams
    ):
        """
        :param target_lens_model: name of target profile
        :param component_lens_model: name of component profile
        :param kwargs_list: list of kwargs of component profile, length of list corresponds to number of components used to fit.
                            The normalization (must be nonzero) will be effectively overridden by the linear weights
        :param lin_fit_hyperparams: kwargs indicating range of fit, number of points to evaluate fit, etc.
        """
        super(SynthesisProfile, self).__init__()
        self.target_class = LensModel([target_lens_model])
        self.component_class = LensModel([component_lens_model])
        self.kwargs_list = kwargs_list
        self.lin_fit_hyperparams = lin_fit_hyperparams
        self.check_num_evals()

    def linear_weight_mle_fit(self, kwargs_target, kwargs_list):
        if self._static is True:
            return self._linear_weights
        else:
            return self._linear_weight_mle_fit(kwargs_target, kwargs_list)

    def _linear_weight_mle_fit(self, kwargs_target, kwargs_list):
        """Fits a linear fit of the amplitudes for each component to minimize a chi2.

        :param kwargs_target: kwargs of target profile to be approximated
        :param kwargs_list: list of kwargs of component profile, length of list
            corresponds to number of components used to fit. The normalization (must be
            nonzero) will be effectively overridden by the linear weights
        """
        self.set_limits(kwargs_list, self.lin_fit_hyperparams)
        kwargs_target_centered = [self.circular_centered_kwargs(kwargs_target[0])]
        Y = self.target_class.kappa(
            x=self.r_eval_list,
            y=np.zeros_like(self.r_eval_list),
            kwargs=kwargs_target_centered,
        )
        M = np.zeros((len(self.r_eval_list), self.num_components))
        C = np.diag(Y * self.sigma)  # covariance matrix between components
        for j in range(
            self.num_components
        ):  # M[i,j] is the jth component 1D kappa evaluated at r[i].
            kwargs = self.circular_centered_kwargs(kwargs_list[j])
            M[:, j] = self.component_class.kappa(
                x=self.r_eval_list, y=0, kwargs=[kwargs]
            )
        MTinvC = np.matmul(M.T, np.linalg.inv(C))
        first_term = np.linalg.inv(np.matmul(MTinvC, M))
        second_term = np.matmul(MTinvC, Y)
        return np.matmul(first_term, second_term)

    def set_static(self, linear_weights):
        """Sets weights to be static self values. Useful to call e.g. function many
        times with the same kwargs. If kwargs_target or kwargs_list change, need to
        rerun linear fit by using set_dynamic.

        :param linear_weights: output of LinearWeightMLEFit
        :return: self weights set
        """
        self._static = True
        self._linear_weights = linear_weights

    def set_dynamic(self):
        self._static = False
        if hasattr(self, "_linear_weights"):
            del self._linear_weights

    def circular_centered_kwargs(self, kwargs):
        """
        :param kwargs: kwargs to remove center and ellipticity for linear fit. These are re-added when functions are called
        """
        kwargs_new = copy.deepcopy(kwargs)
        if "e1" in kwargs_new:
            kwargs_new["e1"] = 0
        if "e2" in kwargs_new:
            kwargs_new["e2"] = 0
        if "center_x" in kwargs_new:
            kwargs_new["center_x"] = 0
        if "center_y" in kwargs_new:
            kwargs_new["center_y"] = 0
        return kwargs_new

    def set_limits(self, kwargs_list, lin_fit_hyperparams):
        """
        :param kwargs_list: list of kwargs of component profile
        :param lin_fit_hyperparams: kwargs indicating range of fit, number of points to evaluate fit, etc.
            'lower_log_bound': log10 innermost radius of fit
            'upper_log_bound': log10 outermost radius of fit
            'num_r_evals': number of locations to evaluate fit to minimize chi2, must be larger than the number of components
            'sigma': used to evaluate chi2. default is 1%
        """
        self.num_components = len(kwargs_list)
        if "lower_log_bound" not in lin_fit_hyperparams:
            self.lower_log_bound = -6
        else:
            self.lower_log_bound = lin_fit_hyperparams["lower_log_bound"]
        if "upper_log_bound" not in lin_fit_hyperparams:
            self.upper_log_bound = 3
        else:
            self.upper_log_bound = lin_fit_hyperparams["upper_log_bound"]
        if "num_r_evals" not in lin_fit_hyperparams:
            self.num_r_evals = 100
        else:
            self.num_r_evals = lin_fit_hyperparams["num_r_evals"]
        if "sigma" not in lin_fit_hyperparams:
            self.sigma = 0.01
        else:
            self.sigma = lin_fit_hyperparams["sigma"]
        self.r_eval_list = np.logspace(
            self.lower_log_bound, self.upper_log_bound, self.num_r_evals
        )

    def function(self, x, y, **kwargs_target):
        """Returns lensing potential.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds) :kwargs_target:
            kwargs of target profile to be approximated
        """
        weight_list = self.linear_weight_mle_fit([kwargs_target], self.kwargs_list)
        f_ = np.zeros_like(x)
        f_innermost = 0  # for some profiles, minimum potential can go below zero. Add a constant here to make zero the minimum
        for kwargs, weight in zip(self.kwargs_list, weight_list):
            f_ += weight * self.component_class.potential(
                x, y, [kwargs]
            )  # -potential_offset)
            f_innermost += weight * self.component_class.potential(
                [10**self.lower_log_bound], [0], [kwargs]
            )
        return f_ - f_innermost

    def derivatives(self, x, y, **kwargs_target):
        """Returns df/dx and df/dy of the function which are the deflection angles.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds) :kwargs_target:
            kwargs of target profile to be approximated
        """
        weight_list = self.linear_weight_mle_fit([kwargs_target], self.kwargs_list)
        f_x, f_y = np.zeros_like(x), np.zeros_like(y)
        for kwargs, weight in zip(self.kwargs_list, weight_list):
            f_x_, f_y_ = weight * np.array(self.component_class.alpha(x, y, [kwargs]))
            f_x += f_x_
            f_y += f_y_
        return f_x, f_y

    def hessian(self, x, y, **kwargs_target):
        """Returns Hessian matrix of function d^2f/dx^2, d^f/dy^2, d^2/dxdy.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds) :kwargs_target:
            kwargs of target profile to be approximated
        """
        weight_list = self.linear_weight_mle_fit([kwargs_target], self.kwargs_list)
        f_xx, f_xy, f_yx, f_yy = (
            np.zeros_like(x),
            np.zeros_like(x),
            np.zeros_like(x),
            np.zeros_like(x),
        )
        for kwargs, weight in zip(self.kwargs_list, weight_list):
            f_xx_i, f_xy_i, f_yx_i, f_yy_i = weight * np.array(
                self.component_class.hessian(x, y, [kwargs])
            )
            f_xx += f_xx_i
            f_xy += f_xy_i
            f_yx += f_yx_i
            f_yy += f_yy_i
        return f_xx, f_xy, f_yx, f_yy

    def check_num_evals(self):
        """Confirm that the number of evaluations is more than the number of components.

        Still not guaranteed to prevent overfitting
        """
        num_comp = len(self.kwargs_list)
        if num_comp >= self.lin_fit_hyperparams["num_r_evals"]:
            raise ValueError(
                "There must be more num_r_evals than components or the profile will be overfit"
            )



from lenstronomy.LightModel.Profiles.nie import NIE
from lenstronomy.LensModel.Profiles.chameleon import Chameleon as ChameleonLens
from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class Chameleon(object):
    """Class of the Chameleon model (See Dutton+ 2011, Suyu+2014) an elliptical
    truncated double isothermal profile."""

    param_names = ["amp", "w_c", "w_t", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "w_c": 0,
        "w_t": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "w_c": 100,
        "w_t": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.nie = NIE()
        self._chameleonLens = ChameleonLens()

    def function(self, x, y, amp, w_c, w_t, e1, e2, center_x=0, center_y=0):
        """

        :param x: ra-coordinate
        :param y: dec-coordinate
        :param w_c:
        :param w_t:
        :param amp: amplitude of first power-law flux
        :param e1: eccentricity parameter
        :param e2: eccentricity parameter
        :param center_x: center
        :param center_y: center
        :return: flux of chameleon profile
        """
        amp_new, w_c, w_t, s_scale_1, s_scale_2 = self._chameleonLens.param_convert(
            amp, w_c, w_t, e1, e2
        )
        flux1 = self.nie.function(x, y, 1, e1, e2, s_scale_1, center_x, center_y)
        flux2 = self.nie.function(x, y, 1, e1, e2, s_scale_2, center_x, center_y)
        flux = amp_new * (flux1 - flux2)
        return flux

    def light_3d(self, r, amp, w_c, w_t, e1, e2, center_x=0, center_y=0):
        """

        :param r: 3d radius
        :param w_c:
        :param w_t:
        :param amp: amplitude of first power-law flux
        :param e1: eccentricity parameter
        :param e2: eccentricity parameter
        :param center_x: center
        :param center_y: center
        :return: 3d flux of chameleon profile at radius r
        """
        amp_new, w_c, w_t, s_scale_1, s_scale_2 = self._chameleonLens.param_convert(
            amp, w_c, w_t, e1, e2
        )
        flux1 = self.nie.light_3d(r, 1, e1, e2, s_scale_1, center_x, center_y)
        flux2 = self.nie.light_3d(r, 1, e1, e2, s_scale_2, center_x, center_y)
        flux = amp_new * (flux1 - flux2)
        return flux


@export
class DoubleChameleon(object):
    """Class of the double Chameleon model.

    See Dutton+2011, Suyu+2014 for the single Chameleon model.
    """

    param_names = [
        "amp",
        "ratio",
        "w_c1",
        "w_t1",
        "e11",
        "e21",
        "w_c2",
        "w_t2",
        "e12",
        "e22",
        "center_x",
        "center_y",
    ]
    lower_limit_default = {
        "amp": 0,
        "ratio": 0,
        "w_c1": 0,
        "w_t1": 0,
        "e11": -0.8,
        "e21": -0.8,
        "w_c2": 0,
        "w_t2": 0,
        "e12": -0.8,
        "e22": -0.8,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "ratio": 100,
        "w_c1": 100,
        "w_t1": 100,
        "e11": 0.8,
        "e21": 0.8,
        "w_c2": 100,
        "w_t2": 100,
        "e12": 0.8,
        "e22": 0.8,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.chameleon = Chameleon()

    def function(
        self,
        x,
        y,
        amp,
        ratio,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        center_x=0,
        center_y=0,
    ):
        """

        :param x:
        :param y:
        :param amp:
        :param ratio:
        :param w_c1:
        :param w_t1:
        :param e11:
        :param e21:
        :param w_c2:
        :param w_t2:
        :param e12:
        :param e22:
        :param center_x:
        :param center_y:
        :return:
        """
        f_1 = self.chameleon.function(
            x, y, amp / (1.0 + 1.0 / ratio), w_c1, w_t1, e11, e21, center_x, center_y
        )
        f_2 = self.chameleon.function(
            x, y, amp / (1.0 + ratio), w_c2, w_t2, e12, e22, center_x, center_y
        )
        return f_1 + f_2

    def light_3d(
        self,
        r,
        amp,
        ratio,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        center_x=0,
        center_y=0,
    ):
        """

        :param r: 3d radius
        :param amp:
        :param ratio: ratio of first to second amplitude of Chameleon surface brightness
        :param w_c1:
        :param w_t1:
        :param e11:
        :param e21:
        :param w_c2:
        :param w_t2:
        :param e12:
        :param e22:
        :param center_x:
        :param center_y:
        :return: 3d light density at radius r
        """
        f_1 = self.chameleon.light_3d(
            r, amp / (1.0 + 1.0 / ratio), w_c1, w_t1, e11, e21, center_x, center_y
        )
        f_2 = self.chameleon.light_3d(
            r, amp / (1.0 + ratio), w_c2, w_t2, e12, e22, center_x, center_y
        )
        return f_1 + f_2


@export
class TripleChameleon(object):
    """Class of the Chameleon model (See Suyu+2014) an elliptical truncated double
    isothermal profile."""

    param_names = [
        "amp",
        "ratio12",
        "ratio13",
        "w_c1",
        "w_t1",
        "e11",
        "e21",
        "w_c2",
        "w_t2",
        "e12",
        "e22",
        "w_c3",
        "w_t3",
        "e13",
        "e23",
        "center_x",
        "center_y",
    ]
    lower_limit_default = {
        "amp": 0,
        "ratio12": 0,
        "ratio13": 0.0,
        "w_c1": 0,
        "w_t1": 0,
        "e11": -0.8,
        "e21": -0.8,
        "w_c2": 0,
        "w_t2": 0,
        "e12": -0.8,
        "e22": -0.8,
        "w_c3": 0,
        "w_t3": 0,
        "e13": -0.8,
        "e23": -0.8,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "ratio12": 100,
        "ratio13": 100,
        "w_c1": 100,
        "w_t1": 100,
        "e11": 0.8,
        "e21": 0.8,
        "w_c2": 100,
        "w_t2": 100,
        "e12": 0.8,
        "e22": 0.8,
        "w_c3": 100,
        "w_t3": 100,
        "e13": 0.8,
        "e23": 0.8,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.chameleon = Chameleon()

    def function(
        self,
        x,
        y,
        amp,
        ratio12,
        ratio13,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        w_c3,
        w_t3,
        e13,
        e23,
        center_x=0,
        center_y=0,
    ):
        """

        :param x:
        :param y:
        :param amp:
        :param ratio12: ratio of first to second amplitude
        :param ratio13: ratio of first to third amplitude
        :param w_c1:
        :param w_t1:
        :param e11:
        :param e21:
        :param w_c2:
        :param w_t2:
        :param e12:
        :param e22:
        :param w_c3:
        :param w_t3:
        :param e13:
        :param e23:
        :param center_x:
        :param center_y:
        :return:
        """
        amp1 = amp / (1.0 + 1.0 / ratio12 + 1.0 / ratio13)
        amp2 = amp1 / ratio12
        amp3 = amp1 / ratio13
        f_1 = self.chameleon.function(
            x, y, amp1, w_c1, w_t1, e11, e21, center_x, center_y
        )
        f_2 = self.chameleon.function(
            x, y, amp2, w_c2, w_t2, e12, e22, center_x, center_y
        )
        f_3 = self.chameleon.function(
            x, y, amp3, w_c3, w_t3, e13, e23, center_x, center_y
        )
        return f_1 + f_2 + f_3

    def light_3d(
        self,
        r,
        amp,
        ratio12,
        ratio13,
        w_c1,
        w_t1,
        e11,
        e21,
        w_c2,
        w_t2,
        e12,
        e22,
        w_c3,
        w_t3,
        e13,
        e23,
        center_x=0,
        center_y=0,
    ):
        """

        :param r: 3d light radius
        :param amp:
        :param ratio12: ratio of first to second amplitude
        :param ratio13: ratio of first to third amplitude
        :param w_c1:
        :param w_t1:
        :param e11:
        :param e21:
        :param w_c2:
        :param w_t2:
        :param e12:
        :param e22:
        :param w_c3:
        :param w_t3:
        :param e13:
        :param e23:
        :param center_x:
        :param center_y:
        :return:
        """
        amp1 = amp / (1.0 + 1.0 / ratio12 + 1.0 / ratio13)
        amp2 = amp1 / ratio12
        amp3 = amp1 / ratio13
        f_1 = self.chameleon.light_3d(r, amp1, w_c1, w_t1, e11, e21, center_x, center_y)
        f_2 = self.chameleon.light_3d(r, amp2, w_c2, w_t2, e12, e22, center_x, center_y)
        f_3 = self.chameleon.light_3d(r, amp3, w_c3, w_t3, e13, e23, center_x, center_y)
        return f_1 + f_2 + f_3



__author__ = "sibirrer"

import copy

import pytest
import numpy.testing as npt
import numpy as np
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.Workflow.fitting_sequence import FittingSequence
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF


class TestFittingSequence(object):
    """Test the fitting sequences."""

    def setup_method(self):
        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 10  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification

        self.kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        data_class = ImageData(**self.kwargs_data)
        kwargs_psf_gaussian = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "pixel_size": deltaPix,
            "truncation": 3,
        }
        psf_gaussian = PSF(**kwargs_psf_gaussian)
        self.kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": psf_gaussian.kernel_point_source,
            "psf_error_map": np.zeros_like(psf_gaussian.kernel_point_source),
        }
        psf_class = PSF(**self.kwargs_psf)
        # 'EXTERNAL_SHEAR': external shear
        kwargs_shear = {
            "gamma1": 0.01,
            "gamma2": 0.01,
        }  # gamma_ext: shear strength, psi_ext: shear angle (in radian)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": 1.8,
            "center_x": 0,
            "center_y": 0,
            "e1": 0.1,
            "e2": 0.1,
        }

        lens_model_list = ["SPEP", "SHEAR"]
        self.kwargs_lens = [kwargs_spemd, kwargs_shear]
        lens_model_class = LensModel(lens_model_list=lens_model_list)
        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 3,
            "center_x": 0,
            "center_y": 0,
            "e1": 0.1,
            "e2": 0.1,
        }

        lens_light_model_list = ["SERSIC"]
        self.kwargs_lens_light = [kwargs_sersic]
        lens_light_model_class = LightModel(light_model_list=lens_light_model_list)
        source_model_list = ["SERSIC_ELLIPSE"]
        self.kwargs_source = [kwargs_sersic_ellipse]
        source_model_class = LightModel(light_model_list=source_model_list)
        self.kwargs_ps = [
            {"ra_source": 0.0, "dec_source": 0.0, "source_amp": 1.0}
        ]  # quasar point source position in the source plane and intrinsic brightness
        point_source_list = ["SOURCE_POSITION"]
        point_source_class = PointSource(
            point_source_type_list=point_source_list, fixed_magnification_list=[True]
        )
        kwargs_numerics = {
            "supersampling_factor": 1,
            "supersampling_convolution": False,
            "compute_mode": "regular",
            "point_source_supersampling_factor": 1,
        }
        imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )
        image_sim = sim_util.simulate_simple(
            imageModel,
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
        )

        data_class.update_data(image_sim)
        self.data_class = data_class
        self.psf_class = psf_class
        self.kwargs_data["image_data"] = image_sim
        self.kwargs_model = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": source_model_list,
            "lens_light_model_list": lens_light_model_list,
            "point_source_model_list": point_source_list,
            "fixed_magnification_list": [False],
            "index_lens_model_list": [[0, 1]],
            "point_source_frame_list": [[0]],
        }
        self.kwargs_numerics = kwargs_numerics

        num_source_model = len(source_model_list)

        self.kwargs_constraints = {
            "num_point_source_list": [4],
            "image_plane_source_list": [False] * num_source_model,
            "solver_type": "NONE",  # 'PROFILE', 'PROFILE_SHEAR', 'ELLIPSE', 'CENTER'
        }

        self.kwargs_likelihood = {
            "force_no_add_image": True,
            "source_marg": True,
            "linear_prior": [1],
            "image_position_uncertainty": 0.004,
            "check_matched_source_position": False,
            "source_position_tolerance": 0.001,
            "source_position_sigma": 0.001,
            "check_positive_flux": True,
        }

        lens_sigma = [
            {
                "theta_E": 0.1,
                "gamma": 0.1,
                "e1": 0.1,
                "e2": 0.1,
                "center_x": 0.1,
                "center_y": 0.1,
            },
            {"gamma1": 0.1, "gamma2": 0.1},
        ]
        lens_lower = [
            {
                "theta_E": 0.0,
                "gamma": 1.5,
                "center_x": -2,
                "center_y": -2,
                "e1": -0.4,
                "e2": -0.4,
            },
            {"gamma1": -0.3, "gamma2": -0.3},
        ]
        lens_upper = [
            {
                "theta_E": 10.0,
                "gamma": 2.5,
                "center_x": 2,
                "center_y": 2,
                "e1": 0.4,
                "e2": 0.4,
            },
            {"gamma1": 0.3, "gamma2": 0.3},
        ]
        source_sigma = [
            {
                "R_sersic": 0.05,
                "n_sersic": 0.5,
                "center_x": 0.1,
                "center_y": 0.1,
                "e1": 0.1,
                "e2": 0.1,
            }
        ]
        source_lower = [
            {
                "R_sersic": 0.01,
                "n_sersic": 0.5,
                "center_x": -2,
                "center_y": -2,
                "e1": -0.4,
                "e2": -0.4,
            }
        ]
        source_upper = [
            {
                "R_sersic": 10,
                "n_sersic": 5.5,
                "center_x": 2,
                "center_y": 2,
                "e1": 0.4,
                "e2": 0.4,
            }
        ]

        lens_light_sigma = [
            {"R_sersic": 0.05, "n_sersic": 0.5, "center_x": 0.1, "center_y": 0.1}
        ]
        lens_light_lower = [
            {"R_sersic": 0.01, "n_sersic": 0.5, "center_x": -2, "center_y": -2}
        ]
        lens_light_upper = [
            {"R_sersic": 10, "n_sersic": 5.5, "center_x": 2, "center_y": 2}
        ]
        ps_sigma = [{"ra_source": 1, "dec_source": 1, "point_amp": 1}]

        lens_param = (
            self.kwargs_lens,
            lens_sigma,
            [{}, {"ra_0": 0, "dec_0": 0}],
            lens_lower,
            lens_upper,
        )
        source_param = (
            self.kwargs_source,
            source_sigma,
            [{}],
            source_lower,
            source_upper,
        )
        lens_light_param = (
            self.kwargs_lens_light,
            lens_light_sigma,
            [{"center_x": 0}],
            lens_light_lower,
            lens_light_upper,
        )
        ps_param = self.kwargs_ps, ps_sigma, [{}], self.kwargs_ps, self.kwargs_ps

        self.kwargs_params = {
            "lens_model": lens_param,
            "source_model": source_param,
            "lens_light_model": lens_light_param,
            "point_source_model": ps_param,
            # 'special': special_param
        }
        image_band = [self.kwargs_data, self.kwargs_psf, self.kwargs_numerics]
        multi_band_list = [image_band]
        self.kwargs_data_joint = {
            "multi_band_list": multi_band_list,
            "multi_band_type": "multi-linear",
        }

    def test_simulationAPI_image(self):
        npt.assert_almost_equal(self.data_class.data[4, 4], 0.1, decimal=0)

    def test_simulationAPI_psf(self):
        npt.assert_almost_equal(
            np.sum(self.psf_class.kernel_point_source), 1, decimal=6
        )

    def test_fitting_sequence(self):
        fittingSequence = FittingSequence(
            self.kwargs_data_joint,
            self.kwargs_model,
            self.kwargs_constraints,
            self.kwargs_likelihood,
            self.kwargs_params,
        )

        kwargs_result = fittingSequence.best_fit(bijective=False)
        lens_temp = kwargs_result["kwargs_lens"]
        npt.assert_almost_equal(
            lens_temp[0]["theta_E"], self.kwargs_lens[0]["theta_E"], decimal=2
        )

        logL = fittingSequence.best_fit_likelihood
        print(logL, "test")
        # print(lens_temp, source_temp, lens_light_temp, ps_temp, special_temp)
        assert logL < 0
        bic = fittingSequence.bic
        assert bic > 0
        # npt.assert_almost_equal(bic, 20000000220.29376, decimal=-4)

        # npt.assert_almost_equal(logL, -10000000061.792593, decimal=-4)

        n_p = 2
        n_i = 2
        fitting_list = []

        kwargs_pso = {"sigma_scale": 1, "n_particles": n_p, "n_iterations": n_i}
        fitting_list.append(["PSO", kwargs_pso])
        kwargs_align = {"delta_shift": 0.2, "n_particles": 2, "n_iterations": 2}
        fitting_list.append(["align_images", kwargs_align])
        kwargs_psf_iter = {
            "num_iter": 2,
            "psf_iter_factor": 0.5,
            "stacking_method": "mean",
            "new_procedure": False,
        }
        fitting_list.append(["psf_iteration", kwargs_psf_iter])
        fitting_list.append(["restart", None])
        fitting_list.append(["fix_not_computed", {"free_bands": [True]}])
        n_sersic_overwrite = 4
        kwargs_update = {
            "lens_light_add_fixed": [[0, ["n_sersic"], [n_sersic_overwrite]]],
            "lens_light_remove_fixed": [[0, ["center_x"]]],
            "change_source_lower_limit": [[0, ["n_sersic"], [0.1]]],
            "change_source_upper_limit": [[0, ["n_sersic"], [10]]],
        }
        fitting_list.append(["update_settings", kwargs_update])

        chain_list = fittingSequence.fit_sequence(fitting_list)
        (
            lens_fixed,
            source_fixed,
            lens_light_fixed,
            ps_fixed,
            special_fixed,
            extinction_fixed,
        ) = fittingSequence._updateManager.fixed_kwargs
        kwargs_result = fittingSequence.best_fit(bijective=False)
        npt.assert_almost_equal(
            kwargs_result["kwargs_lens"][0]["theta_E"],
            self.kwargs_lens[0]["theta_E"],
            decimal=1,
        )
        npt.assert_almost_equal(
            fittingSequence._updateManager._lens_light_fixed[0]["n_sersic"],
            n_sersic_overwrite,
            decimal=8,
        )
        npt.assert_almost_equal(lens_light_fixed[0]["n_sersic"], 4, decimal=-1)
        assert fittingSequence._updateManager._lower_kwargs[1][0]["n_sersic"] == 0.1
        assert fittingSequence._updateManager._upper_kwargs[1][0]["n_sersic"] == 10

        # test 'set_param_value' fitting sequence
        fitting_list = [
            ["set_param_value", {"lens": [[1, ["gamma1"], [0.013]]]}],
            ["set_param_value", {"lens_light": [[0, ["center_x"], [0.009]]]}],
            ["set_param_value", {"source": [[0, ["n_sersic"], [2.993]]]}],
            ["set_param_value", {"ps": [[0, ["ra_source"], [0.007]]]}],
        ]

        fittingSequence.fit_sequence(fitting_list)

        kwargs_set = fittingSequence._updateManager.parameter_state
        assert kwargs_set["kwargs_lens"][1]["gamma1"] == 0.013
        assert kwargs_set["kwargs_lens_light"][0]["center_x"] == 0.009
        assert kwargs_set["kwargs_source"][0]["n_sersic"] == 2.993
        assert kwargs_set["kwargs_ps"][0]["ra_source"] == 0.007

        from unittest import TestCase

        t = TestCase()
        with t.assertRaises(ValueError):
            fitting_list_two = []
            fitting_list_two.append(["fake_mcmc_method", kwargs_pso])
            fittingSequence.fit_sequence(fitting_list_two)

    def test_cobaya(self):
        np.random.seed(42)

        # make a basic lens model to fit
        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 10  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification
        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        data_class = ImageData(**kwargs_data)
        kwargs_psf_gaussian = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "pixel_size": deltaPix,
            "truncation": 3,
        }
        psf_gaussian = PSF(**kwargs_psf_gaussian)

        # make a lens
        lens_model_list = ["SIS"]
        kwargs_lens = [{"theta_E": 1.5, "center_x": 0.0, "center_y": 0.0}]
        lens_model_class = LensModel(lens_model_list=lens_model_list)

        # make a source
        source_model_list = ["SERSIC"]
        kwargs_source = [
            {
                "amp": 1.0,
                "R_sersic": 0.3,
                "n_sersic": 3.0,
                "center_x": 0.1,
                "center_y": 0.1,
            }
        ]
        source_model_class = LightModel(light_model_list=source_model_list)

        kwargs_numerics = {
            "supersampling_factor": 1,
            "supersampling_convolution": False,
        }

        imageModel = ImageModel(
            data_class,
            psf_gaussian,
            lens_model_class,
            source_model_class,
            kwargs_numerics=kwargs_numerics,
        )
        image_sim = sim_util.simulate_simple(imageModel, kwargs_lens, kwargs_source)

        data_class.update_data(image_sim)

        kwargs_data["image_data"] = image_sim

        kwargs_model = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": source_model_list,
        }

        lens_fixed = [{"center_x": 0.0, "center_y": 0.0}]
        lens_sigma = [{"theta_E": 0.01}]
        lens_lower = [{"theta_E": 0.1}]
        lens_upper = [{"theta_E": 3.0}]

        source_fixed = [{}]
        source_sigma = [
            {"R_sersic": 0.01, "n_sersic": 0.01, "center_x": 0.01, "center_y": 0.01}
        ]
        source_lower = [
            {"R_sersic": 0.01, "n_sersic": 0.5, "center_x": -1, "center_y": -1}
        ]
        source_upper = [
            {"R_sersic": 1.0, "n_sersic": 6.0, "center_x": 1, "center_y": 1}
        ]

        lens_param = [kwargs_lens, lens_sigma, lens_fixed, lens_lower, lens_upper]
        source_param = [
            kwargs_source,
            source_sigma,
            source_fixed,
            source_lower,
            source_upper,
        ]

        kwargs_params = {"lens_model": lens_param, "source_model": source_param}

        kwargs_constraints = {}

        multi_band_list = [[kwargs_data, kwargs_psf_gaussian, kwargs_numerics]]

        kwargs_data_joint = {
            "multi_band_list": multi_band_list,
            "multi_band_type": "multi-linear",
        }

        kwargs_likelihood = {"source_marg": False}

        fittingSequence = FittingSequence(
            kwargs_data_joint,
            kwargs_model,
            kwargs_constraints,
            kwargs_likelihood,
            kwargs_params,
        )

        kwargs_cobaya = {
            "proposal_widths": [0.001, 0.001, 0.001, 0.001, 0.001],
            "Rminus1_stop": 100,  # does this need to be large? can we run in test mode?
            "force_overwrite": True,
        }

        chain_list = fittingSequence.fit_sequence(
            [["metropolis_hastings", kwargs_cobaya]]
        )

    def test_zeus(self):
        np.random.seed(42)
        # we make a very basic lens+source model to feed to check zeus can be run through fitting sequence
        # we don't use the kwargs defined in setup() as those are modified during the tests; using unique kwargs here is safer

        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 10  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification

        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        data_class = ImageData(**kwargs_data)
        kwargs_psf_gaussian = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "pixel_size": deltaPix,
            "truncation": 3,
        }
        psf_gaussian = PSF(**kwargs_psf_gaussian)

        # make a lens
        lens_model_list = ["EPL"]
        kwargs_epl = {
            "theta_E": 0.6,
            "gamma": 2.6,
            "center_x": 0.0,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.1,
        }
        kwargs_lens = [kwargs_epl]
        lens_model_class = LensModel(lens_model_list=lens_model_list)

        # make a source
        source_model_list = ["SERSIC_ELLIPSE"]
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 3,
            "center_x": 0.0,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.1,
        }
        kwargs_source = [kwargs_sersic_ellipse]
        source_model_class = LightModel(light_model_list=source_model_list)

        kwargs_numerics = {
            "supersampling_factor": 1,
            "supersampling_convolution": False,
        }

        imageModel = ImageModel(
            data_class,
            psf_gaussian,
            lens_model_class,
            source_model_class,
            kwargs_numerics=kwargs_numerics,
        )
        image_sim = sim_util.simulate_simple(imageModel, kwargs_lens, kwargs_source)

        data_class.update_data(image_sim)

        kwargs_data["image_data"] = image_sim

        kwargs_model = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": source_model_list,
        }

        lens_fixed = [{}]
        lens_sigma = [
            {
                "theta_E": 0.1,
                "gamma": 0.1,
                "e1": 0.1,
                "e2": 0.1,
                "center_x": 0.1,
                "center_y": 0.1,
            }
        ]
        lens_lower = [
            {
                "theta_E": 0.0,
                "gamma": 1.5,
                "center_x": -2,
                "center_y": -2,
                "e1": -0.4,
                "e2": -0.4,
            }
        ]
        lens_upper = [
            {
                "theta_E": 10.0,
                "gamma": 2.5,
                "center_x": 2,
                "center_y": 2,
                "e1": 0.4,
                "e2": 0.4,
            }
        ]

        source_fixed = [{}]
        source_sigma = [
            {
                "R_sersic": 0.05,
                "n_sersic": 0.5,
                "center_x": 0.1,
                "center_y": 0.1,
                "e1": 0.1,
                "e2": 0.1,
            }
        ]
        source_lower = [
            {
                "R_sersic": 0.01,
                "n_sersic": 0.5,
                "center_x": -2,
                "center_y": -2,
                "e1": -0.4,
                "e2": -0.4,
            }
        ]
        source_upper = [
            {
                "R_sersic": 10,
                "n_sersic": 5.5,
                "center_x": 2,
                "center_y": 2,
                "e1": 0.4,
                "e2": 0.4,
            }
        ]

        lens_param = [kwargs_lens, lens_sigma, lens_fixed, lens_lower, lens_upper]
        source_param = [
            kwargs_source,
            source_sigma,
            source_fixed,
            source_lower,
            source_upper,
        ]

        kwargs_params = {"lens_model": lens_param, "source_model": source_param}

        kwargs_constraints = {}

        multi_band_list = [[kwargs_data, kwargs_psf_gaussian, kwargs_numerics]]

        kwargs_data_joint = {
            "multi_band_list": multi_band_list,
            "multi_band_type": "multi-linear",
        }

        kwargs_likelihood = {"source_marg": False}

        fittingSequence = FittingSequence(
            kwargs_data_joint,
            kwargs_model,
            kwargs_constraints,
            kwargs_likelihood,
            kwargs_params,
        )

        fitting_list = []
        kwargs_zeus = {
            "sampler_type": "ZEUS",
            "n_burn": 2,
            "n_run": 2,
            "walkerRatio": 4,
            "backend_filename": "test_mcmc_zeus.h5",
        }

        fitting_list.append(["MCMC", kwargs_zeus])

        chain_list = fittingSequence.fit_sequence(fitting_list)

    def test_multinest(self):
        # Nested sampler tests
        # further decrease the parameter space for nested samplers to run faster

        fittingSequence = FittingSequence(
            self.kwargs_data_joint,
            self.kwargs_model,
            self.kwargs_constraints,
            self.kwargs_likelihood,
            self.kwargs_params,
        )
        fitting_list = []
        kwargs_update = {
            "ps_add_fixed": [[0, ["ra_source", "dec_source"], [0, 0]]],
            "lens_light_add_fixed": [
                [0, ["n_sersic", "R_sersic", "center_x", "center_y"], [4, 0.1, 0, 0]]
            ],
            "source_add_fixed": [
                [
                    0,
                    ["R_sersic", "e1", "e2", "center_x", "center_y"],
                    [0.6, 0.1, 0.1, 0, 0],
                ]
            ],
            "lens_add_fixed": [
                [
                    0,
                    ["gamma", "theta_E", "e1", "e2", "center_x", "center_y"],
                    [1.8, 1.0, 0.1, 0.1, 0, 0],
                ],
                [1, ["gamma1", "gamma2"], [0.01, 0.01]],
            ],
            "change_source_lower_limit": [[0, ["n_sersic"], [2.9]]],
            "change_source_upper_limit": [[0, ["n_sersic"], [3.1]]],
        }
        fitting_list.append(["update_settings", kwargs_update])
        kwargs_multinest = {
            "sampler_type": "MULTINEST",
            "kwargs_run": {
                "n_live_points": 10,
                "evidence_tolerance": 0.5,
                "sampling_efficiency": 0.8,  # 1 for posterior-only, 0 for evidence-only
                "importance_nested_sampling": False,
                "multimodal": True,
                "const_efficiency_mode": False,  # reduce sampling_efficiency to 5% when True
            },
            "remove_output_dir": True,
        }
        fitting_list.append(["nested_sampling", kwargs_multinest])

        chain_list2 = fittingSequence.fit_sequence(fitting_list)
        kwargs_fixed = fittingSequence._updateManager.fixed_kwargs
        npt.assert_almost_equal(kwargs_fixed[0][1]["gamma1"], 0.01, decimal=2)
        assert fittingSequence._updateManager._lower_kwargs[1][0]["n_sersic"] == 2.9
        assert fittingSequence._updateManager._upper_kwargs[1][0]["n_sersic"] == 3.1

        kwargs_test = {"kwargs_lens": 1}
        fittingSequence.update_state(kwargs_test)
        kwargs_out = fittingSequence.best_fit(bijective=True)
        assert kwargs_out["kwargs_lens"] == 1

    def test_dynesty(self):
        np.random.seed(42)
        kwargs_params = copy.deepcopy(self.kwargs_params)
        kwargs_params["lens_model"][0][0]["theta_E"] += 0.01
        kwargs_params["lens_model"][0][0]["gamma"] += 0.01
        fittingSequence = FittingSequence(
            self.kwargs_data_joint,
            self.kwargs_model,
            self.kwargs_constraints,
            self.kwargs_likelihood,
            kwargs_params,
        )

        fitting_list = []
        kwargs_dynesty = {
            "sampler_type": "DYNESTY",
            "kwargs_run": {
                "dlogz_init": 0.01,
                "nlive_init": 20,
                "nlive_batch": 20,
                "maxbatch": 1,
            },
        }

        fitting_list.append(["nested_sampling", kwargs_dynesty])
        chain_list = fittingSequence.fit_sequence(fitting_list)

    def test_nautilus(self):
        np.random.seed(42)
        kwargs_params = copy.deepcopy(self.kwargs_params)
        fittingSequence = FittingSequence(
            self.kwargs_data_joint,
            self.kwargs_model,
            self.kwargs_constraints,
            self.kwargs_likelihood,
            kwargs_params,
        )

        fitting_list = []
        kwargs_nautilus = {
            "prior_type": "uniform",
            "verbose": True,
            "f_live": 1.0,
            "n_eff": 0.0,
            "n_live": 2,
            "seed": 42,
        }

        fitting_list.append(["Nautilus", kwargs_nautilus])
        chain_list = fittingSequence.fit_sequence(fitting_list)

    def test_dypolychord(self):
        fittingSequence = FittingSequence(
            self.kwargs_data_joint,
            self.kwargs_model,
            self.kwargs_constraints,
            self.kwargs_likelihood,
            self.kwargs_params,
        )
        fitting_list = []
        kwargs_dypolychord = {
            "sampler_type": "DYPOLYCHORD",
            "kwargs_run": {
                "ninit": 8,
                "nlive_const": 10,
                #'seed_increment': 1,
                "resume_dyn_run": False,
                #'init_step': 10,
            },
            "polychord_settings": {
                "seed": 1,
                #'num_repeats': 20
            },
            "dypolychord_dynamic_goal": 0.8,  # 1 for posterior-only, 0 for evidence-only
            "remove_output_dir": True,
        }
        fitting_list.append(["nested_sampling", kwargs_dypolychord])
        chain_list = fittingSequence.fit_sequence(fitting_list)

    def test_minimizer(self):
        n_p = 2
        n_i = 2

        fitting_list = []
        kwargs_simplex = {"n_iterations": n_i, "method": "Nelder-Mead"}
        fitting_list.append(["SIMPLEX", kwargs_simplex])
        kwargs_simplex = {"n_iterations": n_i, "method": "Powell"}
        fitting_list.append(["SIMPLEX", kwargs_simplex])
        kwargs_pso = {"sigma_scale": 1, "n_particles": n_p, "n_iterations": n_i}
        fitting_list.append(["PSO", kwargs_pso])
        kwargs_mcmc = {
            "sigma_scale": 1,
            "n_burn": 1,
            "n_run": 1,
            "n_walkers": 10,
            "sampler_type": "EMCEE",
        }
        fitting_list.append(["MCMC", kwargs_mcmc])
        kwargs_mcmc["re_use_samples"] = True
        kwargs_mcmc["init_samples"] = np.array(
            [[np.random.normal(1, 0.001)] for i in range(100)]
        )
        fitting_list.append(["MCMC", kwargs_mcmc])

        def custom_likelihood(
            kwargs_lens,
            kwargs_source=None,
            kwargs_lens_light=None,
            kwargs_ps=None,
            kwargs_special=None,
            kwargs_extinction=None,
        ):
            theta_E = kwargs_lens[0]["theta_E"]
            return -((theta_E - 1.0) ** 2) / 0.1**2 / 2

        kwargs_likelihood = {"custom_logL_addition": custom_likelihood}

        kwargs_data_joint = {"multi_band_list": []}
        kwargs_model = {"lens_model_list": ["SIS"]}
        kwargs_constraints = {}
        lens_param = (
            [{"theta_E": 1, "center_x": 0, "center_y": 0}],
            [{"theta_E": 0.1, "center_x": 0.1, "center_y": 0.1}],
            [{"center_x": 0, "center_y": 0}],
            [{"theta_E": 0, "center_x": -10, "center_y": -10}],
            [{"theta_E": 10, "center_x": 10, "center_y": 10}],
        )

        kwargs_params = {"lens_model": lens_param}
        fittingSequence = FittingSequence(
            kwargs_data_joint,
            kwargs_model,
            kwargs_constraints,
            kwargs_likelihood,
            kwargs_params,
        )
        args = fittingSequence.param_class.kwargs2args(
            kwargs_lens=[{"theta_E": 1, "center_x": 0, "center_y": 0}]
        )
        kwargs_result = fittingSequence.param_class.args2kwargs(args)
        print(kwargs_result)
        print(args, "test args")
        chain_list = fittingSequence.fit_sequence(fitting_list)
        kwargs_result = fittingSequence.best_fit(bijective=False)
        npt.assert_almost_equal(
            kwargs_result["kwargs_lens"][0]["theta_E"], 1, decimal=2
        )


if __name__ == "__main__":
    pytest.main()
